from Fortuna import FlexCat, TruffleShuffle, d


fabric = TruffleShuffle(('Silk', 'Linen', 'Cotton', 'Wool', 'Leather'))

basic_gear = FlexCat({
    "Simple Melee Weapon": (
        "Dagger", "Club", "Hand Axe", "Light Hammer", "Mace",
        "Javelin", "Spear", "Quarterstaff", "Greatclub", "Sickle",
    ),
    "Simple Ranged Weapon": (
        "Light Crossbow", "Dart", "Sling", "Shortbow",
    ),
    "Martial Melee Weapon": (
        "Long Sword", "War Pick", "Morningstar", "Flail", "Battleaxe",
        "Warhammer", "Lance", "Trident", "Greatsword", "Greataxe", "Glaive",
        "Halberd", "Maul", "Pike", "Whip", "Scimitar", "Shortsword", "Rapier",
        "Bastard Sword",
    ),
    "Martial Ranged Weapon": (
        "Heavy Crossbow", "Hand Crossbow", "Blowgun", "Longbow",
    ),
    "Costume": (
        lambda: f"{fabric()} Lingerie",
        "Fancy Costume",  "Nun Costume", "Animal Costume", "Monster Costume",  "Maid Costume",
        "Vampire Costume", "Beholder Costume", "Ninja Costume", "Pirate Costume", "Royal Knight Costume",
    ),
    "Clothing": (
        lambda: f"{fabric()} Robes",
        lambda: f"{fabric()} Hat",
        lambda: f"{fabric()} Gloves",
        lambda: f"{fabric()} Shirt",
        lambda: f"{fabric()} Pants",
        lambda: f"{fabric()} Boots",
        lambda: f"{fabric()} Overcoat",
        "Noble Garb", "Royal Robes", "Villager Garb", "Rag Shirt", "Ball Gown", "Highborn Garb",
        "Kilt", "Apron, Boots, Gloves and Goggles", "Crude Hide", "Leather Chaps", "Royal Garb",
        "Fancy Dress", "Peasant Garb",
    ),
    "Light Armor": (
        "Padded Gambeson", "Leather Jerkin", "Bone Splint Tunic", "Studded Doublet",
    ),
    "Medium Armor": (
        "Chain Mail Shirt", "Ring Mail Tunic", "Scale Mail Suit", "Brigandine",
    ),
    "Heavy Armor": (
        "Plate Mail:\n    Breastplate, Gambeson",
        "Quarter Plate:\n    Breastplate, Gambeson, Gauntlets, Sabatons",
        "Half Plate:\n    Breastplate, Gambeson, Visor Helm"
        "\n    Pauldrons, Gauntlets, Sabatons",
        "Field Plate:\n    Breastplate, Gambeson, Visor Helm"
        "\n    Gorget, Pauldrons, Vambraces, "
        "\n    Gauntlets, Cuisses Leggings, Sabatons",
    ),
    "Shield": (
        "Tower Shield", "Kite Shield", "Jousting Shield", "Heater Shield", "Round Shield", "Buckler",
    ),
    "Ammunition": (
        lambda: f"Quiver of {d(3) * 10} Arrows",
        lambda: f"Quiver of {d(3) * 10} Bolts",
        lambda: f"Bag of {d(3) * 10} Sling Stones",
        lambda: f"Box of {d(3) * 10} Blowgun Darts",
    )
}, key_bias="truffle_shuffle", val_bias="truffle_shuffle")

villager_weapons = TruffleShuffle((
    "Bone Cleaver", "Bowie Knife", "Butcher Knife", "Timber Axe", "Jeweler's Hammer", "Sledge Hammer", "Smith's Hammer",
    "Hunting Spear", "Herb Knife", "Bullwhip", "Pitchfork", "Garden Shovel", "Meat Tenderizer", "Bent Lead Pipe",
    "Matching Bronze Bookends", "Rolling Pin", "Wooden Spoon", "Yard Rake", "Brass Candle Holder", "Camp Axe",
    "Iron Frying Pan", "Bronze Cauldron", "Walking Stick", "Trophy Longsword", "Letter Opener",
    "Broken Axe Handle", "Wooden Stake", "Rusty Shank", "Rake", "Shearing Blade",
))

random_material = TruffleShuffle((
    "Platinum", "Golden", "Silver", "Bronze", "Copper", "Mithral", "Vorpal",
    "Ruby", "Emerald", "Diamond", "Amethyst", "Onyx", "Topaz", "Moonstone", "Sun Stone",
    "Quartz", "Tiger Eye", "Magic", "Broken", "Cursed", "Abyssal",
    "Flame", "Frost", "Storm", "Stone",
))

caster_weapons = TruffleShuffle((  # mage weapons
    "Ceremonial Athame", "Harvester's Boline", "Ritual Scepter", "Gnarled Walking Staff",
    "Garden Sickle", "Incense Brazier", "Ceremonial Spear", "Rune Sword",
    "Reaper Scythe", "Skull on a Chain", "Spirit Totem", "Ritual Doll",
    "Stolen Spell Book", "Iron Bond Tome", "Magical Tome", "Magical Rug",
    lambda: f"{random_material()} Wand",
    lambda: f"{random_material()} Staff",
    lambda: f"{random_material()} Necklace",
    lambda: f"{random_material()} Ring",
    lambda: f"{random_material()} Bottle",
    lambda: f"{random_material()} Sword",
    lambda: f"{random_material()} Chalice",
    lambda: f"{random_material()} Rod",
))

main_hand_weapons = TruffleShuffle((  # one-handed, Fighters
    "Rapier", "Longsword", "Falchion", "Saber", "Pick", "Flail", "Gladius", "Axe", "Hammer", "Mace",
    "Bastard Sword", "Battleaxe", "War Pick", "Morningstar", "Warhammer", "Spear",
))

off_hand_weapons = TruffleShuffle((  # typically dual wielded, Rogues, dex
    "Dagger", "Axe", "Hammer", "Mace", "Sickle", "Scimitar", "Gladius", "Flail", "Sword", "Cat-of-nine-tails",
    "Blackjack", "Club", "Broken Bottle",
))

off_hand_shield = TruffleShuffle((
    "Tower Shield", "Kite Shield", "Jousting Shield", "Heater Shield", "Round Shield", "Buckler", "Duelist Cloak",
))

versatile_weapons = TruffleShuffle((  # typically used with a shield, Fighters
    "Bastard Sword", "Battleaxe", "War Pick", "Morningstar", "Warhammer", "Spear",
))

two_hand_weapons = TruffleShuffle((  # two-handed only, Fighters or Knights
    "Greatsword", "Greataxe", "Great Warhammer", "Executioner's Axe", "Zweihänder", "Claymore",
    "Spiked Greatclub", "Double-sided Scimitar", "War Scythe", "Blade Staff",
))

pole_weapons = TruffleShuffle((  # Knight's lance
    "Lance", "War Spear", "Halberd", "Pike", "Poleaxe", "Trident", "Glaive", "Dragon Lance",
))

thrown_weapons = TruffleShuffle((  # archer main weapons
    "Throwing Knives", "Throwing Axes", "Pilum", "Javelin", "Spear",
))

bow_weapons = TruffleShuffle((
    "Heavy Crossbow", "Longbow", "Elven Bow", "Pair of Hand-crossbows", "Scorpion Batista",
))

favorite_str_weapon = TruffleShuffle((
    pole_weapons,
    two_hand_weapons,
    versatile_weapons,
    main_hand_weapons,
))

favorite_dex_weapon = TruffleShuffle((
    thrown_weapons,
    versatile_weapons,
    main_hand_weapons,
    off_hand_weapons
))

armor_by_ac = {
    5: (
        "Costume", "Noble", "Royal", "Gown", "Dress", "Fancy", "Highborn",
    ),
    8: (
        "Naked", "Lingerie", "Monk", "Rag", "Peasant", "Cotton", "Linen",
    ),
    9: (
        "Villager", "Kilt", "Apron", "Silk", "Wool",
    ),
    10: (
        "Hide", "Leather", "Padded",
    ),
    11: (
        "Jerkin", "Splint", "Studded",
    ),
    12: (
        "Chain", "Ring", "Scale",
    ),
    13: (
        "Plate Mail", "Brigandine",
    ),
    14: (
        "Quarter Plate",
    ),
    15: (
        "Half Plate",
    ),
    16: (
        "Field Plate",
    ),
}

armor_bonus_gear = {
    "Visor Helm", "Gorget", "Pauldrons", "Vambraces", "Gauntlets", "Cuisses Leggings", "Sabatons",
    "Tower Shield", "Kite Shield", "Jousting Shield", "Heater Shield", "Round Shield", "Shield",
    "Buckler", "Duelist Cloak",
}


if __name__ == "__main__":
    print(basic_gear.cat_keys)



from Fortuna import TruffleShuffle, QuantumMonty
from DungeonGenerator.player_lib import Fighter, Paladin, Barbarian, Wizard, Sorcerer, Warlock
from DungeonGenerator.player_lib import Cleric, Druid, Bard, Rogue, Ranger, Monk


random_human = QuantumMonty((
    "Human: Peasant",
    "Human: Villager",
    "Human: Highborn",
    "Human: Outcast",
    "Human: Noble",
    "Human: Outlaw",
    "Human: Royal",
)).front_gauss

npc_dict = {
    'farm_professions': (
        "Bean Farmer", "Grain Farmer", "Vegetable Farmer", "Fruit Farmer", "Pig Farmer",
        "Rancher", "Stable Hand", "Horse Trainer", "Cook", "Ranch Hand", "Rat Catcher",
        "Vagabond", "Weed Farmer", "Bee Keeper", "Worm Farmer", "Gong Farmer", "Maid",
        "Footman", "Servant", "Beggar", "Piss Boy", "Sin Eater", "Town Crier", "Grave Digger",
        "Ditch Digger", "Thug", 'Gypsy', "Nerf Herder", 'Messenger', 'Midwife', 'Milkmaid', 'Miner', 'Minstrel',
        'Peddler', 'Student',
    ),
    'quests': (
        'Perform a rescue mission',
        'Recover a legendary artifact',
        'Remove a powerful curse',
        'Defend a castle from invasion',
        'Spy on a villain',
        'Lay siege to an enemy castle',
        'Restore the family name',
        'Investigate a murder',
        'Solve a riddle',
        'Stop an execution',
        'Locate a missing family member',
        'Break someone out of jail',
        'Preform an ancient ritual to save an endangered species',
        'Communicate with the dead',
        'Recover stolen artwork or heirloom',
        'Give safe passage',
    ),
    'npc_classes': (
        Fighter, Paladin, Barbarian,
        Wizard,
        Sorcerer, Warlock, Bard,
        Cleric, Druid,
        Rogue, Ranger, Monk,
    ),
    'races': (
        random_human, "Mountain Dwarf", "Hill Dwarf", "Shout Halfling",
        "Lightfoot Halfling", "Forest Gnome", "Rock Gnome", "High Elf", "Wood Elf", "Half-elf", "Drow", "Half-orc",
        "Tiefling", "Dragonborn",
    ),
    'appearances': (
        "Distinctive jewelry", "Piercings", "Flamboyant or outlandish clothes", "Formal, clean clothes",
        "Ragged, dirty clothes", "Pronounced scar", "Missing teeth", "Missing fingers", "Unusual eye color", "Tattoos",
        "Birthmark", "Unusual skin color", "Bald", "Braided beard or hair", "Unusual hair color", "Nervous eye twitch",
        "Distinctive nose", "Exceptionally ugly", "Distinctive posture", "Exceptionally beautiful",
    ),
    'talents': (
        "Plays a musical instrument", "Speaks several languages fluently", "Unbelievably lucky",
        "Perfect memory",
        "Great with animals", "Great with children", "Great at solving puzzles", "Great at games",
        "Great at impersonations", "Draws beautifully", "Paints beautifully", "Sings beautifully",
        "Drinks everyone under the table", "Expert carpenter", "Expert cook",
        "Expert dart thrower and rock skipper", "Expert juggler", "Skilled actor and master of disguise",
        "Skilled dancer", "Knows thieves' cant",
    ),
    'mannerisms': (
        "Prone to singing", "Prone to whistling", "Prone to humming quietly",
        "Speaks in rhyme", "Speaks in some peculiar way",
        "Particularly low booming voice", "High-pitched screechy voice", "Slurs words",
        "Enunciates overly clearly", "Speaks with a lisp", "Speaks with a stutter",
        "Speaks loudly", "Whispers", "Uses flowery speech", "Uses unnecessarily long words",
        "Frequently uses the wrong word", "Uses colorful oaths and exclamations", "Makes constant jokes or puns",
        "Prone to predictions of doom", "Fidgets", "Squints", "Stares into the distance", "Chews something",
        "Paces", "Taps fingers", "Bites fingernails", "Twirls hair or tugs beard",
    ),
    'traits': (
        "Argumentative", "Arrogant", "Blustering", "Rude", "Curious", "Friendly", "Brutally Honest", "Hot tempered",
        "Irritable", "Ponderous", "Quiet", "Suspicious", "Nervous",
    ),
    'skills': {
        'STR': ("Athletics STR", "Heavy Lifting STR"),
        'DEX': ("Acrobatics DEX", "Sleight of Hand DEX", "Stealth DEX", "Reflex DEX"),
        'INT': ("Arcana INT", "History INT", "Investigation INT", "Nature INT", "Religion INT",
                "Inspect Merchandise INT"),
        'WIS': ("Animal Handling WIS", "Insight WIS", "Medicine WIS", "Perception WIS", "Survival WIS"),
        'CHA': ("Deception CHA", "Intimidation CHA", "Performance CHA", "Persuasion CHA"),
        'CON': ("Endurance CON", "Fortitude CON")
    },
    'bonds': (
        "Dedicated to fulfilling a personal life goal", "Protective of close family members",
        "Protective of colleagues or compatriots", "Loyal to a benefactor, patron, or employer",
        "Captivated by a romantic interest", "Drawn to a special place", "Protective of a sentimental keepsake",
        "Protective of a valuable possession", "Out for revenge",
    ),
    'flaws': (
        "Forbidden love or susceptibility to romance", "Enjoys decadent pleasures", "Arrogance",
        "Envies another creature's possessions or station", "Overpowering greed", "Prone to rage",
        "Has a powerful enemy", "Specific phobia", "Shameful or scandalous history", "Secret crime or misdeed",
        "Knowledge of forbidden lore", "Foolhardy bravery",
    ),
    'trinkets': (
        'a dead scarab beetle', 'a toy soldier with a missing head', 'a small box filled with buttons',
        "a candle that can't be lit", 'a tiny cage with no door', 'an old key', 'an indecipherable treasure map',
        'a hilt from a broken sword', "a rabbit's foot", 'a glass eye', 'a cameo in the likeness of a famous person',
        'a silver skull the size of a coin', 'an alabaster mask', 'a pyramid of black incense',
        'a nightcap of pleasant dreams', 'a single caltrop made from bone', 'a gold monocle frame without the lens',
        'a 1-inch cube, each side painted a different color', 'a crystal knob from a door',
        'a small packet filled with pink dust', 'a fragment of a beautiful song, written on parchment',
        'a silver teardrop earring made from a real teardrop', 'the shell of an egg painted with wildlife scenes',
        'a fan that, when unfolded, shows a sleeping cat', 'a set of bone bagpipes',
        'a four-leaf clover pressed inside a book', 'a drawing of a complex mechanical contraption',
        'an ornate scabbard that fits no blade you have found so far',
        'an invitation to a party where a murder happened',
        'a bronze pentacle with an etching of an animal in its center',
        'a purple handkerchief embroidered with the name zarn the great and powerful',
        'a floor plan for a temple, castle, or some other structure',
        'a bit of folded cloth that, when unfolded, turns into a stylish cap',
        'a receipt of deposit at a bank in a far-flung city', 'a diary with seven missing pages',
        'an empty silver snuffbox', 'an iron holy symbol devoted to an unknown god',
        "a book that tells the story of a legendary hero", 'a vial of dragon blood',
        'an ancient arrow of elven design', 'a needle that never bends', 'an ornate brooch of dwarven design',
        'an empty wine bottle bearing a pretty label', 'a mosaic tile with a multicolored, glazed surface',
        'a petrified mouse', "a black pirate flag adorned with a dragon's skull and crossbones",
        "a tiny mechanical spider that moves about when it's not being observed",
        'a glass jar containing lard with a label that reads, "griffon grease"',
        'a wooden box that holds a living worm with a head on each end of its body',
        'a metal urn containing the ashes of a hero'
    ),
    'professions': (
        'Acrobat', 'Animal Trainer', 'Armor Smith', 'Baker', 'Barmaid', 'Barber', 'Bartender',
        'Blacksmith', 'Bookbinder', 'Bowyer Fletcher', 'Brewer', 'Butcher', 'Butler',
        'Candlestick Maker', 'Chef', 'Clerk', 'Jester', 'Courtesan', 'Dancer',
        'Farrier', 'Ferryman', 'Fire Eater', 'Fisherman', 'Fool', 'Fortuneteller', 'Gardener',
        'Gatekeeper', 'Glassblower', 'Hatter', 'Healer', 'Herbalist', 'Innkeeper',
        'Jailer', 'Juggler', 'Lamp Wright', 'Leather Worker', 'Locksmith', 'Lumberjack',
        'Merchant', 'Missionary', 'Mortician', 'Musician', 'Oracle', 'Playwright', 'Poet',
        'Potion Salesman', 'Potter', 'Prostitute', 'Quartermaster',
        'Sailor', 'Scholar', 'Scout', 'Scribe', 'Sculptor', 'Sentry',
        'Ship Captain', 'Shipwright', 'Silver Smith', 'Singer', 'Soldier', 'Songwriter', 'Spice Dealer',
        'Stage Hand', 'Stonemason', 'Storyteller',
        'Tailor', 'Teacher', 'Tracker', 'Trapper', 'Weaver', 'Wet Nurse', 'Wine Vintner', "Weapon Smith",
    ),
    'noble professions': (
        'Actor', 'Artist', 'Alchemist', 'Artificer', 'Astrologer', 'Author', 'Bureaucrat', 'Coin Maker', 'Diplomat',
        'Falconer', 'Gem Cutter', 'Gold Smith', 'Harpist', 'Instrument Maker', 'Inventor', 'Mad Scientist',
        'Professor', 'Council Member', 'Royal Guard', 'Time Keeper', 'Tutor', 'Local Magistrate', 'Jeweler',
    ),
    'treasure': (
        "Alchemy jug", "Bowl of commanding water rand_elemental", "Brazier of commanding fire rand_elemental",
        "Candle of invocation", "Censer of controlling air rand_elemental", "Dust of disappearance",
        "Dust of dryness", "Dust of sneezing and choking", "Efreeti bottle", "Elixir of health",
        "Eversmoking bottle", "Figurine of wondrous power (Bronze griffon)",
        "Figurine of wondrous power (Ebonyfly)", "Figurine of wondrous power (Golden lions)",
        "Figurine of wondrous power (Ivorygoats)", "Figurine of wondrous power (Marble elephant)",
        "Figurine of wondrous power (Onyx dog)", "Figurine of wondrous power (Serpentine owl)",
        "Figurine of wondrous power (obsidian steed)", "Figurine of wondrous power (silver raven)",
        "Stone of controlling earth rand_elemental", "Stone of good luck",
        "Deck of illusions", "Manual of bodily health", "Manual of gainful exercise",
        "Manual of golems", "Manual of quickness of action", "Scroll of protection",
        "Tome of clear thought",
        "Tome of leadership and influence", "Tome of the stilled tongue", "Tome of understanding",
        "Berserker axe", "Dagger of venom", "Dancing sword", "Defender",
        "Dragon slayer", "Dwarven thrower", "Flame tongue", "Frost brand", "Giant slayer",
        "Hammer of thunderbolts", "Holy avenger", "Immovable rod", "Javelin of lightning", "Luck blade",
        "Mace of smiting", "Mace of terror", "Nine lives stealer", "Oath bow", "Quiver of Ehlonna",
        "Rod of absorption", "Rod of alertness", "Rod of lordly might", "Rod of resurrection",
        "Rod of rulership", "Rod of security", "Rod of the pact keeper", "Scimitar of speed",
        "Sphere of annihilation", "Staff of charming", "Staff of fire", "Staff of frost", "Staff of healing",
        "Staff of power", "Staff of striking", "Staff of swarming insects", "Staff of the adder",
        "Staff of the magi", "Staff of the python", "Staff of the woodlands", "Staff of thunder and lightning",
        "Staff of withering", "Sun blade", "Sword of answering", "Sword of life stealing",
        "Sword of sharpness",
        "Sword of vengeance", "Sword of wounding", "Tentacle rod", "Trident of fish command", "Vicious weapon",
        "Vorpal sword", "Wand of binding", "Wand of enemy detection", "Wand of fear", "Wand of fireballs",
        "Wand of lightning bolts", "Wand of magic detection", "Wand of magic missiles", "Wand of paralysis",
        "Wand of polymorph", "Wand of secrets", "Wand of the war mage", "Wand of web", "Wand of wonder",
        "Weapon of warning", "Animated shield", "Armor of invulnerability",
        "Armor of resistance (breastplate)",
        "Armor of resistance (chain mail)", "Armor of resistance (chain shirt)",
        "Armor of resistance (half plate)", "Armor of resistance (leather)",
        "Armor of resistance (scale mail)",
        "Armor of resistance (splint)", "Armor of resistance (studded leather)", "Armor of vulnerability",
        "Arrow-catching shield", "Belt of cloud giant strength", "Belt of dwarvenkind",
        "Belt of fire giant strength", "Belt of frost (or stone) giant strength",
        "Belt of hill giant strength",
        "Belt of storm giant strength", "Boots of elvenkind", "Boots of levitation", "Boots of speed",
        "Boots of striding and springing", "Boots of the winterlands", "Bracers of archery",
        "Bracers of defense", "Cap of water breathing", "Cape of the mountebank", "Circlet of blasting",
        "Cloak of arachnida", "Cloak of displacement", "Cloak of elvenkind", "Cloak of invisibility",
        "Cloak of protection", "Cloak of the manta ray", "Cloak of the bat", "Demon armor", "Dragon scale mail",
        "Dwarven plate", "Elven chain", "Eyes of charming", "Eyes of minute seeing", "Eyes of the eagle",
        "Gauntlets of ogre power", "Glamoured studded leather", "Gloves of missile snaring",
        "Gloves of swimming and climbing", "Gloves of thievery", "Goggles of night", "Hat of disguise",
        "Headband of intellect", "Helm of brilliance", "Helm of comprehending languages", "Helm of telepathy",
        "Helm of teleportation", "Mantle of spell resistance", "Mariner's armor", "Mithral armor",
        "Plate armor of etherealness", "Plate armor of resistance", "Robe of eyes",
        "Robe of scintillating colors", "Robe of stars", "Robe of the archmagi", "Robe of useful items_by_rank",
        "Sentinel shield", "Shield of missile attraction", "Slippers of spider climbing", "Spellguard shield",
        "Winged boots", "Iron bands of Bilarro", "Rope of climbing", "Saddle of the cavalier",
        "Horseshoes of a zephyr", "Horseshoes of speed", "Apparatus of Kwalish", "Bag of beans",
        "Bag of holding", "Broom of flying", "Carpet of flying", "Chime of opening",
        "Crystal ball (very rare version)", "Crystal ball (legendary version)", "Cube of force", "Cubic gate",
        "Daern's instant fortress", "Decanter of endless water", "Dimensional shackles", "Driftglobe",
        "Efreeti chain", "Folding boat", "Heward's handy haversack", "Horn of Valhalla (bronze)",
        "Horn of Valhalla (iron)", "Horn of Valhalla (silver or brass)", "Horn of blasting",
        "Lantern of revealing", "Mirror of life trapping",
        "Pipes of haunting", "Pipes of the sewers", "Portable hole", "Quaal's feather token",
        "Rope of entanglement", "Sending stones", "Well of many worlds", "Wind fan", "Wings of flying",
        "Amulet of health", "Amulet of proof against detection and location", "Amulet of the planes",
        "Brooch of shielding", "Elemental gem", "Gem of brightness", "Gem of seeing", "Medallion of thoughts",
        "Necklace of adaptation", "Necklace of fireballs", "Necklace of prayer beads", "Pearl of power",
        "Periapt of health", "Periapt of proof against poison", "Periapt of wound closure",
        "Ring of X-ray vision", "Ring of air elemental command", "Ring of animal influence",
        "Ring of djinni summoning", "Ring of earth elemental command", "Ring of evasion",
        "Ring of feather falling", "Ring of fire elemental command", "Ring of free action",
        "Ring of invisibility", "Ring of jumping", "Ring of mind shielding", "Ring of protection",
        "Ring of regeneration", "Ring of resistance", "Ring of shooting stars", "Ring of spell storing",
        "Ring of spell turning", "Ring of swimming", "Ring of telekinesis", "Ring of the ram",
        "Ring of three wishes", "Ring of warmth", "Ring of water elemental command", "Ring of water walking",
        "Ring of Fortuna", "Scarab of protection",
    )
}

favorite_trinket = TruffleShuffle(npc_dict['trinkets'])



from Fortuna import *
from DungeonGenerator.cr import CR
from DungeonGenerator.treasure import Loot
from DungeonGenerator.utilities import cr_str


class Trap:
    name = "Basic"
    type = "Minor"
    spot_dc = 10
    save_dc = 11
    damage = "1d10"
    rank_0_damage = "1d2"
    rank_1_damage = "2d4"
    rank_2_damage = "3d6"
    rank_3_damage = "4d8"
    rank_4_damage = "5d10"
    random_ability = TruffleShuffle(("CON", "STR", "DEX", "INT", "WIS"))
    random_trap = FlexCat({
        'bludgeoning': ("Falling Rocks", "Swinging Hammer", "Falling Chandelier"),
        'falling': ("Deep Pit", "Spring Loaded Floor Tile", "Greased Slide", "False Floor"),
        'piercing': ("Spiked Pit", "Guided Arrow", "Flying Broken Glass", "Falling Chandelier of Spikes"),
        'slashing': ("Swinging Blade", "Guillotine", "Hooks and Chains", "Cloud of Daggers", "Exploding Mirror"),
        'poison': ("Poison Dart", "Noxious Gas", "Pit Vipers", "Exploding Rotten Corpse"),
        'acid': ("Acid Spay", "Acid Pool", "Acidic Bile", "Dripping Acidic Goo"),
        'fire': ("Flame Jet", "Explosive Gas", "Lava Pit", "Delayed Fireball Blast", "Inferno"),
        'lightning': ("Lightning Rod", "Tesla Coil", "Electric Trip Wire", "Lightning Bolt"),
        'cold': ("Frozen Ground", "Cone of Cold", "Hail Storm", "Freeze the Flesh", "Ice Lance"),
        'necrotic': ("Putrid Spores", "Cloud of Decay", "Energy Drain", "Beam of Entropy"),
    }, key_bias="truffle_shuffle", val_bias="truffle_shuffle")
    treasure = Loot()

    def __init__(self, cr, dam_type=None):
        self.cr = smart_clamp(cr, -3, 30)
        self.save_vs = self.random_ability()
        self.xp = self.set_xp_by_cr(self.cr)
        if dam_type:
            self.damage_type = dam_type
        else:
            self.damage_type = self.random_trap.random_cat()
        self.name = self.random_trap(self.damage_type)
        self.damage = self.set_damage(self.cr)

    def set_damage(self, cr):
        if cr < 1:
            damage = self.rank_0_damage
        elif cr < 5:
            damage = self.rank_1_damage
        elif cr < 11:
            damage = self.rank_2_damage
        elif cr < 17:
            damage = self.rank_3_damage
        else:
            damage = self.rank_4_damage
        return damage

    @staticmethod
    def set_xp_by_cr(cr):
        xp_lookup = (
            10, 25, 50, 100, 200, 450, 700, 1100, 1800, 2300, 2900, 3900, 5000, 5900, 7200, 8400, 10000,
            11500, 13000, 15000, 18000, 20000, 22000, 25000, 33000, 41000, 50000, 62000, 155000, 155000,
            155000, 155000, 155000, 155000, 155000, 155000, 155000, 155000, 155000, 155000
        )
        return xp_lookup[cr + 3]

    def __lt__(self, other):
        return self.cr <= other.cr

    def __str__(self):
        self.output = (
            f"{self.type} Trap, CR {cr_str(self.cr)}: {self.name}",
            f"Spot & Disarm DC {self.spot_dc}",
            f"Save vs. {self.save_vs} DC {self.save_dc} for half damage.",
            f"Damage: {self.damage} {self.damage_type}",
            f"Disarm XP: {self.xp}",
            ""
        )
        return "\n".join(self.output)


class MinorTrap(Trap):
    type = "Minor"
    spot_dc = 10
    save_dc = 11
    rank_0_damage = "1d4"
    rank_1_damage = "2d4"
    rank_2_damage = "3d4"
    rank_3_damage = "4d4"
    rank_4_damage = "5d4"


class DangerousTrap(Trap):
    type = "Dangerous"
    spot_dc = 12
    save_dc = 15
    rank_0_damage = "1d6"
    rank_1_damage = "2d6"
    rank_2_damage = "3d6"
    rank_3_damage = "4d6"
    rank_4_damage = "5d6"


class DeadlyTrap(Trap):
    type = "Deadly"
    spot_dc = 16
    save_dc = 20
    rank_0_damage = "1d8"
    rank_1_damage = "2d8"
    rank_2_damage = "3d8"
    rank_3_damage = "4d8"
    rank_4_damage = "5d8"


class EpicTrap(Trap):
    type = "Epic"
    spot_dc = 18
    save_dc = 22
    rank_0_damage = "1d10"
    rank_1_damage = "2d10"
    rank_2_damage = "3d10"
    rank_3_damage = "4d10"
    rank_4_damage = "5d10"


random_trap_by_tier = FlexCat({
    '1': (MinorTrap, DangerousTrap, MinorTrap, DeadlyTrap),
    '2': (DangerousTrap, MinorTrap, DeadlyTrap, EpicTrap),
    '3': (DeadlyTrap, DangerousTrap, EpicTrap, MinorTrap),
    '4': (DeadlyTrap, EpicTrap, DangerousTrap, MinorTrap),
}, key_bias="front_linear", val_bias="front_linear")


def random_trap(cr, dam_type=None, room_name="Any"):
    _ = room_name
    tier = str(CR(cr).tier)
    return random_trap_by_tier(tier, cr, dam_type)


if __name__ == "__main__":
    print()
    for i in range(-3, 31):
        print(random_trap(i))



from Fortuna import *
from DungeonGenerator.dungeon import Dungeon, Wilderness, Settlement
from DungeonGenerator.npc import Npc
from DungeonGenerator.treasure import magic_table_by_cr, get_villain_loot, Loot
from DungeonGenerator.utilities import cr_adapter
from DungeonGenerator.quest_lib import things_dict


class Quest:
    trinkets = TruffleShuffle(things_dict['trinkets'])
    actions = TruffleShuffle(things_dict["actions"])
    skills = TruffleShuffle(things_dict["skills"])

    def __init__(self, cr, dungeon_name=None, dungeon_size=4, wilderness_size=1, settlement_size=1):
        self.cr = smart_clamp(cr, -3, 30)
        self.npc = Npc(self.cr)
        self.race = self.npc.race
        self.appearance = self.npc.appearance
        self.mannerism = self.npc.mannerism
        self.background = self.npc.background
        self.objective = self.trinkets() if percent_true(50) else self.actions()
        self.skill = self.skills()
        self.wilderness = Wilderness(self.cr, num_levels=wilderness_size, areas_per_level=10)
        self.dungeon_name = dungeon_name
        self.dungeon = Dungeon(self.cr + wilderness_size, dungeon_name, dungeon_levels=dungeon_size)
        self.reward = get_villain_loot(self.dungeon.threats[-1].threat.cr)
        self.reward.magic_items.append(magic_table_by_cr(self.cr))
        self.dc = self.dungeon.threats[-1].threat.save_dc
        self.city = Settlement(cr=self.cr, num_people=settlement_size*10)
        self.total_xp = self.dungeon.total_xp + self.wilderness.total_xp
        self.summary = '\n'.join((
            f"NPC: Primary Quest Giver",
            f"Race: {self.race}",
            f"Appearance: {self.appearance}",
            f"Mannerism: {self.mannerism}",
            f"Background: {self.background}",
            f"",
            f"Quest Details:",
            f"Starting Location: {self.city.name}",
            f"Objective: {self.objective}.",
            f"Location: Beyond the {self.wilderness.name}, deep inside the {self.dungeon.name}.",
            f"Primary Villain: {self.dungeon.threats[-1].threat.name}",
            f"Quest Skill Check: {self.skill} {self.dc}",
            f"Reward: {self.reward}",
            f"",
        ))
        self.total_treasure = sum((
            self.dungeon.total_treasure,
            self.wilderness.total_treasure,
            self.city.total_treasure,
            self.reward,
        ), Loot())

    def __repr__(self):
        output = (
            f"{self.summary}",
            f"{self.city}",
            f"{self.wilderness}",
            f"{self.dungeon}",
        )
        return '\n'.join(output)


if __name__ == "__main__":
    print("")
    quest = Quest(
        cr=cr_adapter(num_players=5, average_level=5, difficulty=0),
    )
    print(quest)
    print(quest.dungeon)



from Fortuna import *
from DungeonGenerator.cr import CR
from DungeonGenerator.monsters import Villain, Monster, ItemEncounter
from DungeonGenerator.npc_lib import npc_dict, random_human
from DungeonGenerator.gear import BasicGear, Potion
from DungeonGenerator.skills import SkillSet
from DungeonGenerator.treasure import get_loot, Loot, career_loot
from DungeonGenerator.player_lib import Player, Fighter, Paladin, Barbarian, Wizard, Sorcerer, Warlock, random_character_class
from DungeonGenerator.player_lib import Cleric, Druid, Bard, Rogue, Ranger, Monk, Knight
from DungeonGenerator.backgrounds import random_background


class Npc:
    professions = TruffleShuffle(npc_dict['professions'])
    races = QuantumMonty(npc_dict['races']).front_linear
    appearances = TruffleShuffle(npc_dict['appearances'])
    talents = TruffleShuffle(npc_dict['talents'])
    mannerisms = TruffleShuffle(npc_dict['mannerisms'])
    traits = TruffleShuffle(npc_dict['traits'])
    bonds = TruffleShuffle(npc_dict['bonds'])
    flaws = TruffleShuffle(npc_dict['flaws'])
    trinkets = TruffleShuffle(npc_dict['trinkets'])
    treasure = Loot()
    save_dc_list = (
        13, 13, 13, 13, 13, 13, 13, 14, 15, 15, 15, 16, 16, 16, 17, 17, 18, 18,
        18, 18, 19, 19, 19, 19, 20, 20, 20, 21, 21, 21, 22, 22, 22, 23, 24
    )
    racial_bonuses = {
        "Human": {"STR": 1, 'DEX': 1, 'CON': 1, 'INT': 1, 'WIS': 1, 'CHA': 1},
        "Mountain Dwarf": {"STR": 2, 'CON': 2},
        "Hill Dwarf": {"WIS": 1, 'CON': 2},
        "Shout Halfling": {"DEX": 2, 'CON': 1},
        "Lightfoot Halfling": {"DEX": 2, 'CHA': 1},
        "Forest Gnome": {"INT": 2, 'DEX': 1},
        "Rock Gnome": {"INT": 2, 'CON': 2},
        "High Elf": {"DEX": 2, 'INT': 1},
        "Wood Elf": {"DEX": 2, 'WIS': 1},
        "Half-elf": {"CHA": 2, "DEX": 1, "CON": 1},
        "Drow": {"DEX": 2, 'CHA': 1},
        "Half-orc": {"STR": 2, 'CON': 1},
        "Dragonborn": {"STR": 2, 'CHA': 1},
        "Tiefling": {"CHA": 2, 'INT': 1},
    }

    def __init__(self, cr=1, race=None):
        self.name = "NPC"
        self.damage = 0
        self.cr = smart_clamp(cr, 1, 20)
        self.profession = self.professions()
        self.race = race if race else random_human() if percent_true(50) else self.races()
        self.appearance = self.appearances()
        self.talent = self.talents()
        self.mannerism = self.mannerisms()
        self.trait = self.traits()
        self.bond = self.bonds()
        self.flaw = self.flaws()
        self.trinket = self.trinkets()
        self.save_dc = self.save_dc_list[self.cr + 3]
        self.xp = 0
        self.stats = {}
        self.stat_mod = {}
        self.stat_mod_str = {}
        self._set_stats()
        self._set_racial_bonuses()
        self._set_stat_mods()
        self.background = random_background()
        self.saving_throws = self.set_saves()
        self.health = self._set_health()
        self.ac = 10 + self.stat_mod["DEX"]

    def _set_racial_bonuses(self):
        if self.race in self.racial_bonuses.keys():
            for stat, val in self.racial_bonuses[self.race].items():
                self.stats[stat] += val

    def _set_health(self):
        return smart_clamp(4 + self.stat_mod['CON'], 1, 20)

    def _set_stats(self):
        stats = ('STR', 'INT', 'DEX', 'WIS', 'CON', 'CHA')
        for ability in stats:
            self.stats[ability] = ability_dice(3)

    def _set_stat_mods(self):
        mod_by_stat = (
            -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10
        )
        for ability in self.stats:
            self.stat_mod[ability] = mod_by_stat[smart_clamp(self.stats[ability], 1, 30)]
            if self.stat_mod[ability] > 0:
                self.stat_mod_str[ability] = f"+{self.stat_mod[ability]}"
            else:
                self.stat_mod_str[ability] = f"{self.stat_mod[ability]}"

    def set_saves(self):
        result = []
        for key, val in self.stat_mod.items():
            if val > 0:
                result.append(f"{key} +{val}")
            else:
                result.append(f"{key} {val}")
        return result

    @property
    def ability_scores(self):
        return ", ".join(f"{key} {val} ({self.stat_mod_str[key]})" for key, val in self.stats.items())

    def __repr__(self):
        self.output = (
            f"NPC: {self.profession}",
            f"Race: {self.race}",
            f"Appearance: {self.appearance}",
            f"Mannerism: {self.mannerism}",
            f"{self.ability_scores}",
            f"Hit Points: {self.health}",
            f"Armor Class: {self.ac}",
            f"Background: {self.background}",
            ""
        )
        return "\n".join(self.output)


class QuestGiver(Npc):
    random_quest = TruffleShuffle(npc_dict['quests'])
    skill_list = TruffleShuffle((
        "Acrobatics DEX",
        "Animal handling WIS",
        "Arcana INT",
        "Athletics STR",
        "Deception CHA",
        "History INT",
        "Insight WIS",
        "Intimidation CHA",
        "Investigation INT",
        "Medicine WIS",
        "Nature INT",
        "Perception WIS",
        "Performance CHA",
        "Persuasion CHA",
        "Religion INT",
        "Sleight of hand DEX",
        "Stealth DEX",
        "Survival WIS"
    ))

    def __init__(self, cr, race=None):
        super().__init__(cr, race)
        self.quest = self.random_quest()
        self.magic_item = ItemEncounter(cr).item_name
        self.coinage = get_loot(min(30, self.cr + 5))
        self.skill = self.skill_list()
        self.chest = get_loot(max(-3, self.cr - 15), is_villain=True)

    def __repr__(self):
        self.output = (
            f"NPC: Secondary Quest Giver",
            f"Race: {self.race}",
            f"Appearance: {self.appearance}",
            f"Mannerism: {self.mannerism}",
            f"Background: {self.background}",
            f"Quest: {self.quest}",
            f"Required PC Skill: {self.skill} {self.save_dc}",
            f"Quest Rewards: (the party should choose one below)",
            f"1. Mysterious magic item: {self.magic_item}",
            f"2. Fat bag of coins: {self.coinage}",
            f"3. Gem encrusted treasure chest: {self.chest}",
            ""
        )
        return "\n".join(self.output)


class ShopKeeper(Npc):
    def __init__(self, cr, race=None):
        super().__init__(cr, race)
        self.profession = 'Shop Keeper'
        if cr <= 5:
            self.for_sale = [BasicGear(cr) for _ in range(dice(2, 4))]
        elif cr <= 10:
            self.for_sale = [BasicGear(cr - 4) for _ in range(d(4))]
            self.for_sale.extend([BasicGear(cr) for _ in range(dice(2, 4))])
        elif cr <= 15:
            self.for_sale = [BasicGear(cr - 8) for _ in range(d(4))]
            self.for_sale.extend([BasicGear(cr - 4) for _ in range(d(4))])
            self.for_sale.extend([BasicGear(cr) for _ in range(dice(2, 4))])
        else:
            self.for_sale = [BasicGear(cr - 12) for _ in range(d(4))]
            self.for_sale.extend([BasicGear(cr - 8) for _ in range(d(4))])
            self.for_sale.extend([BasicGear(cr - 4) for _ in range(d(4))])
            self.for_sale.extend([BasicGear(cr) for _ in range(dice(2, 4))])
        self.trinket_price = f"{BasicGear(5).price}"
        self.for_sale_names = [f"  {val.name} ({val.id})" for val in self.for_sale]

    def __repr__(self):
        self.output = (
            f"NPC: {self.profession}",
            f"Race: {self.race}",
            f"Appearance: {self.appearance}",
            f"Inspect Merchandise INT: DC {self.save_dc}",
            f"For Sale:",
            "\n".join(self.for_sale_names),
            f"  Trinket: {self.trinket}, Price {self.trinket_price}",
            ""
        )
        return "\n".join(self.output)


class PotionDealer(Npc):
    def __init__(self, cr, race=None):
        super().__init__(cr, race)
        self.profession = 'Potion Dealer'
        self.for_sale = sorted([f"  {Potion(cr)}" for _ in range(dice(3, 4))])

    def __repr__(self):
        self.output = (
            f"NPC: {self.profession}",
            f"Race: {self.race}",
            f"Mannerism: {self.mannerism}",
            f"For Sale:",
            "\n".join(self.for_sale),
            ""
        )
        return "\n".join(self.output)


class Mercenary(Npc):
    pc_classes = CumulativeWeightedChoice((
        (0, Knight),
        (0, Paladin),
        (0, Druid),
        (0, Monk),
        (20, Fighter),
        (40, Cleric),
        (60, Rogue),
        (70, Bard),
        (80, Ranger),
        (90, Barbarian),
        (95, Wizard),
        (97, Sorcerer),
        (99, Warlock),
    ), flat=False)

    def __init__(self, cr, race=None, pc_class=None):
        super().__init__(cr, race)
        self.profession = "Mercenary"
        if pc_class is None:
            self.npc_class = self.pc_classes()
        else:
            self.npc_class = pc_class
        self.level = smart_clamp(cr - 1, 1, 20)
        self.player = Player(self.npc_class, self.level)
        self.stats = self.player.stats
        self.stat_mod = self.player.stat_mod
        self.fav_stat_mod = self.stat_mod[self.npc_class.fav_stat]
        self.magic_bonus = self.player.magic_bonus
        x = self.player.level
        y = self.class_price_multi(self.player.cls.name)
        z = self.fav_stat_mod + self.magic_bonus
        self.price = smart_clamp(x * y * z, x, 99999)
        self.name = f"NPC {self.player.cls.name}"
        self.damage_bonus = self.stat_mod[self.npc_class.fav_stat] + self.player.magic_bonus
        self.damage_type = self.player.damage_type
        self.special_dam_type = self.player.cls.damage_type
        self.damage_formula = self.player.damage_formula
        self.special_dam = f"{Monster(self.level).damage}"
        self.skills = SkillSet(self.level, self.player.prof_bonus, self.stats, self.background)
        self.inventory = Loot()
        while len(self.inventory.magic_items) < CR(self.cr).tier:
            self.inventory.magic_items = career_loot(1, self.level - 1, self.level).magic_items
        if len(self.inventory.magic_items) > self.level:
            self.inventory.magic_items = self.inventory.magic_items[:self.level]
        self.treasure = Loot()

    @staticmethod
    def class_price_multi(class_name):
        output = {
            "Monk": 1,
            "Barbarian": 2,
            "Fighter": 3,
            "Rogue": 4,
            "Bard": 5,
            "Druid": 6,
            "Ranger": 7,
            "Warlock": 8,
            "Sorcerer": 9,
            "Paladin": 10,
            "Cleric": 11,
            "Wizard": 12,
            "Artificer": 13,
            "Knight": 14,
        }
        for key in output:
            if key in class_name:
                return output[key]

    def find_skills(self):
        skills = set(self.background.skills)
        for ability in self.stats:
            if self.stat_mod[ability] > 0 and percent_true(self.stat_mod[ability] * 20):
                skills.add(f"{random_value(npc_dict['skills'][ability])}")
        return list(skills)

    def __repr__(self):
        coinage = ["Copper", "Silver", "Electrum", "Gold", "Platinum"]
        price_dom = coinage[smart_clamp(round(self.player.level / 5), 0, 4)]
        self.output = (
            f"NPC: {self.profession}",
            f"Salary: {self.price} {price_dom}",
            f"Class: {self.player.cls.name}",
            f"Level: {self.player.level}",
            f"Race: {self.race}",
            f"Appearance: {self.appearance}",
            f"{self.player.ability_scores}",
            f"Proficiency Bonus: {self.player.prof_bonus_str}",
            f"Preferred Weapon: {self.player.cls.pref_weapon}",
            f"Hit Points: {self.player.health}",
            f"Armor Class: {self.player.ac}",
            f"Attack Bonus: {self.player.att_bonus}",
            f"Attack Damage: {self.damage_formula} {self.damage_type}",
            f"Special Attack: {self.special_dam} {self.special_dam_type}",
            f"Save DC: {self.player.save_dc}",
            f"Save Proficiencies: {', '.join(self.player.saves)}",
            f"Save Modifiers: {', '.join(self.player.saving_throws)}",
            f"Skills: {self.skills}",
            f"Talent: {self.talent}",
            f"Background: {self.background}",
            f"Inventory: {self.inventory}",
            ""
        )
        return "\n".join(self.output)


class Adventurer(Mercenary):
    goals = TruffleShuffle(npc_dict['quests'])

    def __init__(self, cr, race=None, pc_class=None):
        pcc = random_character_class() if not pc_class else pc_class
        super().__init__(cr + 1, race, pcc)
        self.goal = self.goals()

    def __repr__(self):
        self.output = (
            f"NPC: Adventurer",
            f"Class: {self.player.cls.name}",
            f"Level: {self.player.level}",
            f"Race: {self.race}",
            f"Appearance: {self.appearance}",
            f"{self.player.ability_scores}",
            f"Proficiency Bonus: {self.player.prof_bonus_str}",
            f"Preferred Weapon: {self.player.cls.pref_weapon}",
            f"Hit Points: {self.player.health}",
            f"Armor Class: {self.player.ac}",
            f"Attack Bonus: {self.player.att_bonus}",
            f"Attack Damage: {self.damage_formula} {self.damage_type}",
            f"Special Attack: {self.special_dam} {self.special_dam_type}",
            f"Save DC: {self.player.save_dc}",
            f"Save Proficiencies: {', '.join(self.player.saves)}",
            f"Save Modifiers: {', '.join(self.player.saving_throws)}",
            f"Skills: {self.skills}",
            f"Talent: {self.talent}",
            f"Personal Goal: {self.goal}",
            f"Background: {self.background}",
            f"Inventory: {self.inventory}",
            ""
        )
        return "\n".join(self.output)


class VillainNPC(Adventurer):

    def __init__(self, cr, race=None, pc_class=None):
        super().__init__(cr, race, pc_class)
        self.name = f"Villain disguised as a {self.professions()}"
        self.treasure = get_loot(self.level, is_villain=True)
        self.xp = Villain(self.cr).xp

    def __repr__(self):
        self.output = (
            f"NPC: {self.name}",
            f"Class: {self.player.cls.name}",
            f"Level: {self.player.level}",
            f"Race: {self.race}",
            f"Appearance: {self.appearance}",
            f"Mannerism: {self.mannerism}",
            f"{self.player.ability_scores}",
            f"Proficiency Bonus: {self.player.prof_bonus_str}",
            f"Preferred Weapon: {self.player.cls.pref_weapon}",
            f"Hit Points: {self.player.health}",
            f"Armor Class: {self.player.ac}",
            f"Attack Bonus: {self.player.att_bonus}",
            f"Attack Damage: {self.damage_formula} {self.damage_type}",
            f"Special Attack: {self.special_dam} {self.special_dam_type}",
            f"Save DC: {self.player.save_dc}",
            f"Save Proficiencies: {', '.join(self.player.saves)}",
            f"Save Modifiers: {', '.join(self.player.saving_throws)}",
            f"Skills: {self.skills}",
            f"Talent: {self.talent}",
            f"Background: {self.background}",
            f"XP Value: {self.xp}",
            f"Inventory: {self.treasure}",
            ""
        )
        return "\n".join(self.output)


class Farmer(Npc):
    professions = TruffleShuffle(npc_dict['farm_professions'])


random_weighted_npc = RelativeWeightedChoice((
    (2, Npc),
    (1, Farmer),
    (1, ShopKeeper),
    (1, PotionDealer),
    (2, QuestGiver),
    (1, Mercenary),
    (1, Adventurer),
    (1, VillainNPC),
))


def random_city_npc(cr=1, room_name="Any"):
    _ = room_name
    return random_weighted_npc(cr)


def random_dungeon_npc(cr=1, room_name="Any"):
    if any(x in room_name.lower() for x in ("Barracks", "Training Room", "Armory")) or percent_true(25):
        return Adventurer(cr)
    else:
        return Npc(cr)


if __name__ == "__main__":
    print()
    print(Adventurer(d(1)))
    print(Mercenary(d(1)))
    print(VillainNPC(d(1)))



from Fortuna import *
from DungeonGenerator.traps import random_trap
from DungeonGenerator.monsters import Monster, MonsterGroup, MinionGroup, Boss, Villain, ItemEncounter
from DungeonGenerator.npc import random_dungeon_npc, random_city_npc
from DungeonGenerator.utilities import cr_adapter
from DungeonGenerator.treasure import magic_table_f


rand_color = TruffleShuffle((
    "blue",
    "red",
    "green",
    "yellow",
    "purple",
    "orange",
    "magenta",
    "indigo",
    "cyan",
    "pink",
    "golden",
))


class RoomJunk:
    rand_direction = TruffleShuffle((
        lambda: f"{'north' if percent_true(75) else 'north eastern'}",
        lambda: f"{'south' if percent_true(75) else 'south western'}",
        lambda: f"{'east' if percent_true(75) else 'south eastern'}",
        lambda: f"{'west' if percent_true(75) else 'north western'}",
    ))

    rand_container = TruffleShuffle((
        lambda: f"{rand_color()} bags",
        lambda: f"{rand_color()} baskets",
        lambda: f"{rand_color()} boxes",
        lambda: f"{rand_color()} crates",
        "bookshelves",
        "treasure chests",
        "coffins",
    ))

    rand_door_cond = QuantumMonty((
        "closed but unlocked",
        "closed and locked",
        "bared from the other side",
        "bared from this side",
        "open and unlocked",
        "magically sealed",
    ))

    rand_troop = TruffleShuffle((
        "Footman",
        "Soldier",
        "Archer",
        "Elite Guard",
        "Officer",
        "Siege Master",
        "War Mage",
        "War Priest",
        "Commander",
        "Knight"
    ))

    bridge_materials = QuantumMonty((
        "rope",
        "wooden",
        "stone",
        "rusty iron",
        "polished bronze",
        "gold wire",
        "petrified wood",
        "solid gold",
    ))


class Room:
    _names = TruffleShuffle((
        lambda: f"{RoomJunk.rand_troop()}'s Armory",
        lambda: f"Hatchery",
        TruffleShuffle(("Religious Library", "Arcane Library", "Mundane Library", "Royal Archives")),
        lambda: f"{RoomJunk.rand_troop()}'s Training Room",
        lambda: f"Treasury",
        lambda: f"Tormented Artist Studio",
        lambda: f"Backstage Dressing Room",
        lambda: f"Haunted Auditorium",
        lambda: f"Abandoned Ballroom",
        lambda: f"Carved Marble Great Hall",
        lambda: f"Gnomish Workshop",
        lambda: f"Goblin Casino",
        lambda: f"Guard Room",
        lambda: f"Bestiary",
        lambda: f"Prison Cell",
        lambda: f"Prison Workout Room",
        lambda: f"Torture Chamber",
        lambda: f"Crypt",
        lambda: f"Combat Pit {'Arena' if percent_true(75) else 'Bleachers'}",
        lambda: f"Shrine of Fortuna",
        lambda: f"Desecrated Altar of Light",
        lambda: f"Unholy Temple",
        lambda: f"Desecrated Chapel",
        lambda: f"Kitchen",
        lambda: f"Dinning Hall",
        lambda: f"{RoomJunk.rand_troop()}'s Barracks",
        lambda: f"Hot Spring Bath",
        lambda: f"Horse Stable",
        lambda: f"Mythical Beast Stable",
        lambda: f"Wizard Spire",
        lambda: f"Wizard's Component Pantry",
        lambda: f"Scribe's Hall of Many Desks",
        lambda: f"Artificer's Workshop",
        lambda: f"Alchemy Lab",
        lambda: f"Iron Forge",
        lambda: f"Chamber of Summoning",
        lambda: f"Lavish Bedroom",
        lambda: f"Sewer Tunnel Junction",
        lambda: f"Arboretum",
        lambda: f"Storage Room",
        lambda: f"Wine Cellar",
        lambda: f"Ritual Chamber",
        lambda: f"Observatory",
        lambda: f"Lavatory",
        lambda: f"Game Room",
        lambda: f"Scientific Laboratory",
        lambda: f"Embalming Room",
        lambda: f"Panic Room",
    ))

    _exits = QuantumMonty((
        lambda: f"{RoomJunk.rand_direction()} hallway to the next room.",
        lambda: f"{RoomJunk.rand_direction()} tunnel to the next room.",
        lambda: f"{RoomJunk.rand_direction()} {RoomJunk.bridge_materials.front_linear()} bridge to the next room.",
        lambda: f"wooden stairway to the next room.",
        lambda: f"metal stairway to the next room.",
        lambda: f"carved stone stairway to the next room.",
        lambda: f"{rand_color()} inlaid tile stairway to the next room.",
        lambda: f"magic portal to the next room.",
    ))

    _doors = QuantumMonty((
        lambda: f"Wooden Door, {RoomJunk.rand_door_cond.front_gauss()}",
        lambda: f"Heavy Wooden Door, {RoomJunk.rand_door_cond.front_linear()}",
        lambda: f"Carved Wooden Door, {RoomJunk.rand_door_cond.front_linear()}",
        lambda: f"Studded Wooden Door, {RoomJunk.rand_door_cond.front_linear()}",
        lambda: f"Stone Archway",
        lambda: f"Bronze Portcullis, {RoomJunk.rand_door_cond.quantum_monty()}",
        lambda: f"Solid Iron Gate, {RoomJunk.rand_door_cond.quantum_monty()}",
        lambda: f"Ornate Iron Gate, {RoomJunk.rand_door_cond.quantum_monty()}",
        lambda: f"Inscribed Wooden Door, {RoomJunk.rand_door_cond.quantum_monty()}",
        lambda: f"Hidden Trapdoor in the floor",
        lambda: f"Hidden Trapdoor in the ceiling",
        lambda: f"Secret Door",
        lambda: f"Gold Inlaid Wooden Door, {RoomJunk.rand_door_cond.back_linear()}",
        lambda: f"Etched Silver Door, {RoomJunk.rand_door_cond.back_linear()}",
        lambda: f"Solid Gold Gateway, {RoomJunk.rand_door_cond.back_gauss()}",
    ))

    _sizes = QuantumMonty((
        lambda: f"rectangle, {d(6) * 12} x {-3 + d(6) * 12} hands",
        lambda: f"square, {d(6) * 12} hands across",
        lambda: f"hexagon, {d(6) * 12} hands across",
        lambda: f"octagon, {d(6) * 12} hands across",
        lambda: f"decagon, {d(6) * 12} hands across",
        lambda: f"dodecagon, {d(8) * 8} hands across",
        lambda: f"circular, {dice(2, 5) * 10} hands across",
    ))

    _details = QuantumMonty((
        lambda: f"This room appears empty except for a thick coat of dust on the floor.",
        lambda: f"This room appears empty except for various broken bones scattered around the room.",
        lambda: f"This room appears empty except for blood dripping from the ceiling.",
        lambda: f"This room appears empty except for scattered bits of broken furniture.",
        lambda: f"Empty {RoomJunk.rand_container()} line the walls of this room.",
        lambda: f"Garbage litters the floor of this room.",
        lambda: f"Garbage litters the floor, a clue about an up-coming villain might be found here: 10% chance.",
        lambda: f"The walls are covered in patches of {rand_color()} mold.",
        lambda: f"The floors are covered in clusters of a {rand_color()} fungus.",
        lambda: f"This room is spotless, a clue about an up-coming villain might be found here: 25% chance.",
        lambda: f"Small woodland creatures live in the piles of garbage scattered throughout this room.",
        lambda: f"This room is under construction, woodworking tools are scattered about.",
        lambda: f"Part of the floor is collapsing under its own weight.",
        lambda: f"Part of the ceiling is collapsing under its own weight.",
        lambda: f"{rand_color().title()} coins are scattered on the floor. "
                f"Upon further inspection one would find them glued in place.",
        lambda: f"The room is lined with magical runes carved into the walls.",
        lambda: f"Everything in this room has been painted a strange {rand_color()} color.",
        lambda: f"This room has a large 'x' painted on the floor in black tar.",
        lambda: f"The walls are covered in mold and mildew, someone has written 'HELP ME' in blood on the far wall.",
        lambda: f"Murky {rand_color()} water, three hands deep, "
                f"covers the floor in this room and the one beyond.",
        lambda: f"Small piles of red dust can be found in this room.",
        lambda: f"The walls of this room are covered in highly advanced mathematical formulae.",
        lambda: f"A diabolic monster once used this room to butcher wayward adventurers.",
        lambda: f"An illusion appears in this room, possibly revealing a clue about the Boss: 25% chance.",
        lambda: f"This room is decorated to look like the inside of a ship at sea.",
        lambda: f"At first this room looks like it was decorated for a grand party, but on closer inspection "
                f"the decorations are actually bits of trash hung around the room by rusty wire.",
        lambda: f"This room was decorated years ago for a party or wedding that never happened.",
        lambda: f"This room has a giant map painted in the center of the floor, "
                f"a significant portion of the map is missing.",
        lambda: f"Murky black water covers the floor in this room. A feral kitten is stranded on floating debris.",
        lambda: f"Mid-century modern with a bi-friendly open relationship floor plan.",
        lambda: f"Empty {RoomJunk.rand_container()} line the walls of this room. "
                f"A narcoleptic rune cat is sleeping in the corner.",
        lambda: f"This room reminds me of grandmother's house... before the 'accident'",
        lambda: f"Empty {RoomJunk.rand_container()} line the walls of this room. "
                f"An invisible tome might be found here: 25% chance.",
        lambda: f"Empty {RoomJunk.rand_container()} line the walls of this room. "
                f"A magic item might be found here: 25% chance. {magic_table_f()}.",
        lambda: f"There are a dozen or more humanoids hanging upside-down from the ceiling, writhing in unison.",
    ))
    random_threat = RelativeWeightedChoice((
        (62, Monster),
        (5, MinionGroup),
        (10, MonsterGroup),
        (3, ItemEncounter),
        (5, Villain),
        (5, random_dungeon_npc),
        (10, random_trap),
    ))
    random_threat_no_traps = RelativeWeightedChoice((
        (62, Monster),
        (5, MinionGroup),
        (10, MonsterGroup),
        (5, Villain),
        (5, random_dungeon_npc),
    ))

    def __init__(self, cr=1, room_name="Any", make_traps=True):
        self.cr = cr
        self.name = self._names() if room_name == "Any" else room_name
        self.detail = self._details.front_linear()
        self.size = self._sizes.front_linear()
        if make_traps:
            self.threat = self.random_threat(self.cr, room_name=self.name)
        else:
            self.threat = self.random_threat_no_traps(self.cr, room_name=self.name)
        self.door = self._doors.front_linear()
        self.exit = self._exits.front_linear()
        if hasattr(self.threat, 'xp'):
            self.xp = self.threat.xp
        else:
            self.xp = 0

    def __repr__(self):
        self.output = (
            f"{self.name}",
            f"Dimensions: {self.size}",
            f"Details: {self.detail}",
            f"{self.threat}",
            f"{self.door}, {self.exit}",
            ""
        )
        return "\n".join(self.output)


class WildArea(Room):
    random_threat = RelativeWeightedChoice((
        (50, Monster),
        (30, MinionGroup),
        (2, MonsterGroup),
        (3, ItemEncounter),
        (1, Villain),
        (10, random_city_npc),
        (1, random_dungeon_npc),
        (1, random_trap),
    ))

    def __repr__(self):
        return f"{self.threat}"


class MinionRoom(Room):
    def __init__(self, cr, room_name="Any", make_traps=True):
        super().__init__(cr, room_name, make_traps)
        self.detail = self._details.front_gauss()
        self.size = self._sizes.front_gauss()
        self.door = self._doors.front_gauss()
        self.exit = self._exits.front_gauss()
        self.threat = MinionGroup(self.cr, self.name)


class VillainRoom(Room):
    def __init__(self, cr, room_name="Any", make_traps=True):
        super().__init__(cr, room_name, make_traps)
        self.detail = self._details.quantum_monty()
        self.size = self._sizes.quantum_monty()
        self.door = self._doors.quantum_monty()
        self.exit = self._exits.quantum_monty()
        self.threat = Villain(self.cr, self.name)


class BossRoom(Room):
    def __init__(self, cr, room_name="Any", make_traps=True):
        super().__init__(cr, room_name, make_traps)
        self.name = "Dungeon Heart" if room_name == "Any" else room_name
        self.detail = self._details.back_linear()
        self.size = self._sizes.back_gauss()
        self.door = self._doors.back_gauss()
        self.exit = self._exits.back_gauss()
        self.threat = Boss(self.cr, self.name)

    def __repr__(self):
        self.output = (
            f"{self.name}",
            f"Dimensions: {self.size}",
            f"Details: {self.detail}",
            f"{self.threat}",
            f"Dungeon Exit: {self.door}, {self.exit.replace('to the next room', 'leads out of the dungeon')} "
        )
        return "\n".join(self.output)


if __name__ == '__main__':
    print()
    staring_cr = cr_adapter(num_players=5, average_level=10, difficulty=2)
    haunted_mansion = (
        MinionRoom(staring_cr, room_name="Entrance"),
        Room(staring_cr),
        Room(staring_cr),
        Room(staring_cr),
        VillainRoom(staring_cr + 1),
        MinionRoom(staring_cr),
        Room(staring_cr),
        Room(staring_cr),
        Room(staring_cr),
        BossRoom(staring_cr + 2),
    )
    for room in haunted_mansion:
        print(room)



from Fortuna import *
from DungeonGenerator.cr import CR
from DungeonGenerator.utilities import damage_types, cr_str
from DungeonGenerator.treasure import Loot, get_loot, magic_table_f, magic_table_g, magic_table_h, magic_table_i, no_coin
from DungeonGenerator.monsters_lib import motivation, random_monster, damage_type_cat


class Monster:

    def __init__(self, cr, room_name="Any", name=None):
        self.name = random_monster('monster') if not name else name
        self.cr = CR(cr)
        self.cr_val = self.cr.val
        self.tier = self.cr.tier
        self.cr_key = self.cr.key
        self.total_hp = self.set_hit_points(self.cr_val)
        self.current_hp = self.total_hp
        self.damage = (
            "1d2", "1d2 + 1", "1d2 + 2", "1d2 + 3",
            "1d6 + 8", "1d6 + 14", "1d6 + 20", "1d6 + 26", "1d6 + 32",
            "1d6 + 40", "1d6 + 45", "1d6 + 50", "1d6 + 65", "1d6 + 70",
            "1d8 + 75", "1d8 + 80", "1d8 + 85", "1d8 + 90", "1d8 + 95",
            "1d10 + 100", "1d10 + 105", "1d10 + 110", "1d10 + 115", "1d10 + 120",
            "1d12 + 140", "1d12 + 150", "1d12 + 160", "1d12 + 180", "1d12 + 200",
            "1d20 + 260", "1d20 + 270", "1d20 + 280", "1d20 + 290", "1d20 + 300",
        )[self.cr_key]
        self.xp = self.set_xp_by_cr(self.cr_val)
        self.motive = motivation()
        self.treasure = get_loot(self.cr_val, monster_name=self.name, room_name=room_name, is_villain=False)
        self.prof_bonus = self.set_prof_bonus(self.cr_val)
        variance = plus_or_minus_gauss(3)
        self.ac = self.set_ac(self.cr_val) + variance
        self.att_bonus = self.set_att_bonus(self.cr_val) + -variance
        self.save_dc = self.set_save_dc(self.cr_val)
        self.num_appearing = 1
        self.dam_type = self.find_dam_type()
        self.at_will_dam = f"{smart_clamp(self.cr_val // 3, 1, 10)}d6 + {self.prof_bonus} {self.dam_type}"
        self.rank_by_cr = tuple(x // 5 for x in range(34))
        self.abilities = ["STR", "DEX", "INT", "WIS", "CHA", "CON"]
        self.rank = self.rank_by_cr[self.cr.key]
        self.stats = {
            ability: ability_dice(4) for ability in self.abilities
        }
        self.stat_mod = {}
        self.stat_mod_str = {}
        self.set_stat_mods()
        self.special_damage = f"{self.damage} {damage_types()}"

    def set_stat_mods(self):
        mod_by_stat = (
            -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10
        )
        for ability in self.stats:
            self.stat_mod[ability] = mod_by_stat[smart_clamp(self.stats[ability], 1, 30)]
            if self.stat_mod[ability] > 0:
                self.stat_mod_str[ability] = f"+{self.stat_mod[ability]}"
            else:
                self.stat_mod_str[ability] = f"{self.stat_mod[ability]}"

    def find_dam_type(self):
        dam_type = damage_types()
        for key, val in damage_type_cat.items():
            if any(ele in self.name for ele in val):
                dam_type = key
                break
        if dam_type in ("Physical", "Arcane", "Unholy", "Holy", "Nature", "Divine"):
            dam_type = damage_types(dam_type)
        return dam_type

    @property
    def ability_scores(self):
        return ", ".join(f"{key} {val} ({self.stat_mod_str[key]})" for key, val in self.stats.items())

    def __lt__(self, other):
        return self.cr_val < other.cr

    def __repr__(self):
        self.output = (
            f"Monster: {self.name}, CR {cr_str(self.cr_val)}",
            f"{self.ability_scores}",
            f"Hit Points: {self.total_hp}",
            f"Armor Class: {self.ac}",
            f"Attack Bonus: {self.att_bonus}",
            f"Basic Attack: {self.at_will_dam}",
            f"Save DC: {self.save_dc}",
            f"XP Value: {self.xp}",
            f"Treasure: {self.treasure}",
            ""
        )
        return '\n'.join(self.output)

    @staticmethod
    def cr_mod(num_players, average_level, difficulty):
        cr_modifier = (-3, -3, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
        diff_mod = {'Easy': -1, 'Normal': 0, 'Hard': 1, 'Deadly': 2, 'Insane': 3}
        return average_level + cr_modifier[num_players + diff_mod[difficulty]]

    @staticmethod
    def set_prof_bonus(cr):
        stats_by_cr = (
            2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5,
            5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 9, 9, 10
        )
        return stats_by_cr[cr + 3]

    @staticmethod
    def set_ac(cr):
        stats_by_cr = (
            10, 11, 12, 13, 13, 13, 14, 14, 14, 15, 15, 16, 16, 17, 17, 17, 18,
            18, 18, 18, 19, 19, 19, 19, 19, 20, 20, 20, 20, 20, 21, 21, 21, 22,
        )
        return stats_by_cr[cr + 3]

    @staticmethod
    def set_hit_points(cr):
        stats_by_cr = (
            (1, 6), (7, 35), (36, 49), (50, 70), (71, 85), (86, 100),
            (101, 115), (116, 130), (131, 145), (146, 160), (161, 175),
            (176, 190), (191, 205), (206, 220), (221, 235), (236, 250),
            (251, 265), (266, 280), (281, 295), (296, 310), (311, 325),
            (326, 340), (341, 355), (356, 400), (401, 445), (446, 490),
            (491, 535), (536, 580), (581, 625), (626, 670), (671, 715),
            (716, 760), (761, 805), (806, 850), (851, 900)
        )
        return random_range(stats_by_cr[cr + 3][0], stats_by_cr[cr + 3][1] + 1)

    @staticmethod
    def set_att_bonus(cr):
        stats_by_cr = (
            3, 3, 3, 3, 3, 3, 4, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8,
            8, 9, 10, 10, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 14, 15
        )
        return stats_by_cr[cr + 3]

    @staticmethod
    def set_damage(cr):
        stats_by_cr = (
            (1, 2), (2, 3), (4, 5), (6, 8),
            (9, 14), (15, 20), (21, 26), (27, 32), (33, 38),
            (39, 44), (45, 50), (51, 56), (57, 62), (63, 68),
            (69, 74), (75, 80), (81, 86), (87, 92), (93, 98),
            (99, 104), (105, 110), (111, 116), (117, 122), (123, 140),
            (141, 158), (159, 176), (177, 194), (195, 212), (213, 230),
            (231, 248), (249, 266), (267, 284), (285, 302), (303, 320), (321, 350)
        )
        return random_range(stats_by_cr[cr + 3][0], stats_by_cr[cr + 3][1] + 1)

    @staticmethod
    def set_save_dc(cr):
        stats_by_cr = (
            13, 13, 13, 13, 13, 13, 13, 14, 15, 15, 15, 16, 16, 16, 17, 17, 18, 18,
            18, 18, 19, 19, 19, 19, 20, 20, 20, 21, 21, 21, 22, 22, 22, 23, 24
        )
        return stats_by_cr[cr + 3]

    @staticmethod
    def set_xp_by_cr(cr):
        xp_lookup = (
            10, 25, 50, 100, 200, 450, 700, 1100, 1800, 2300, 2900, 3900, 5000, 5900, 7200, 8400, 10000,
            11500, 13000, 15000, 18000, 20000, 22000, 25000, 33000, 41000, 50000, 62000, 155000, 155000,
            155000, 155000, 155000, 155000, 155000, 155000, 155000, 155000, 155000, 155000
        )
        return xp_lookup[cr + 3]


class MonsterGroup(Monster):
    def __init__(self, cr, room_name="Any", name=None):
        super().__init__(cr, room_name, name)
        self.num_appearing = dice(2, 2)
        self.individual_hp = max(1, int(self.total_hp // self.num_appearing))
        for _ in range(self.num_appearing - 1):
            self.treasure += get_loot(self.cr_val, self.name, room_name)
        self.damage_dice = (
            "1d2", "1d2", "1d2", "1d2",
            "1d6", "1d6", "1d6", "1d6", "1d6",
            "1d6", "1d6", "1d6", "1d6", "1d6",
            "1d8", "1d8", "1d8", "1d8", "1d8",
            "1d10", "1d10", "1d10", "1d10", "1d10",
            "1d12", "1d12", "1d12", "1d12", "1d12",
            "1d20", "1d20", "1d20", "1d20", "1d20"
        )[self.cr_val + 3]
        self.damage_bonus = (
            1, 1, 2, 3,
            8, 14, 20, 26, 32,
            40, 45, 50, 65, 70,
            75, 80, 85, 90, 95,
            100, 105, 110, 115, 120,
            140, 150, 160, 180, 200,
            260, 270, 280, 290, 300
        )[self.cr_val + 3]
        actual_damage_bonus = self.damage_bonus // self.num_appearing
        if actual_damage_bonus > 0:
            self.individual_dam = f"{self.damage_dice} + {actual_damage_bonus}"
        else:
            self.individual_dam = f"{self.damage_dice}"
        self.at_will_dam = f"{smart_clamp(self.cr_val // 4, 1, 5)}d6 + {self.prof_bonus} {self.dam_type}"

    def __repr__(self):
        self.output = (
            f"Monster Group: {self.name}, CR {cr_str(self.cr_val)}",
            f"Number Appearing: {self.num_appearing}",
            f"{self.ability_scores}",
            f"Hit Points: {self.individual_hp} each",
            f"Armor Class: {self.ac}",
            f"Attack Bonus: {self.att_bonus}",
            f"Attack Damage: {self.num_appearing}x ({self.at_will_dam})",
            f"Save DC: {self.save_dc}",
            f"XP Value: {self.xp}",
            f"Treasure: {self.treasure}",
            ""
        )
        return '\n'.join(self.output)


class MinionGroup(Monster):
    def __init__(self, cr, room_name="Any", name=None):
        minion_cr = cr - 3 if cr < 6 else cr // 2
        minion_cr = smart_clamp(minion_cr, -3, 17)
        minion_name = name or random_monster("minion")
        super().__init__(minion_cr, room_name, minion_name)
        self.dam_type = self.find_dam_type()
        self.num_appearing = dice(2, 2)
        self.stats = {
            ability: ability_dice(7) for ability in self.abilities
        }
        self.set_stat_mods()
        self.individual_hp = max(1, int(self.total_hp // self.num_appearing))
        self.treasure = Loot()
        self.damage_dice = (
            "1d2", "1d2", "1d4", "1d4",
            "1d6", "1d6", "1d6", "1d6", "1d6",
            "1d6", "1d6", "1d6", "1d6", "1d6",
            "1d8", "1d8", "1d8", "1d8", "1d8",
            "1d10", "1d10", "1d10", "1d10", "1d10",
            "1d12", "1d12", "1d12", "1d12", "1d12",
            "1d20", "1d20", "1d20", "1d20", "1d20"
        )[minion_cr + 3]
        self.damage_bonus = (
            0, 2, 4, 6,
            8, 14, 20, 26, 32,
            40, 45, 50, 65, 70,
            75, 80, 85, 90, 95,
            100, 105, 110, 115, 120,
            140, 150, 160, 180, 200,
            260, 270, 280, 290, 300
        )[minion_cr + 3]
        actual_damage_bonus = self.damage_bonus // self.num_appearing
        if actual_damage_bonus > 0:
            self.individual_dam = f"{self.damage_dice} + {actual_damage_bonus}"
        else:
            self.individual_dam = f"{self.damage_dice}"
        self.at_will_dam = f"{smart_clamp(self.cr_val // 5, 1, 5)}d4 + {self.prof_bonus} {self.dam_type}"

    def __repr__(self):
        self.output = (
            f"Minion Group: {self.name}, CR {cr_str(self.cr_val)}",
            f"Number Appearing: {self.num_appearing}",
            f"{self.ability_scores}",
            f"Hit Points: {self.individual_hp} each",
            f"Armor Class: {self.ac}",
            f"Attack Bonus: {self.att_bonus}",
            f"Attack Damage: {self.num_appearing}x ({self.at_will_dam})",
            f"Save DC: {self.save_dc}",
            f"XP Value: {self.xp}",
            ""
        )
        return '\n'.join(self.output)


class ItemEncounter(Monster):
    items_by_rank = (
        magic_table_f,
        magic_table_g,
        magic_table_h,
        magic_table_i,
    )

    def __init__(self, cr, room_name="Any", name=""):
        cr_rand = cr + plus_or_minus_linear(2)
        self.cr = smart_clamp(cr_rand + 3 if "legendary" in name.lower() else cr_rand, -3, 30)
        item_rank = smart_clamp(self.cr + 3 // 5, 0, 3)
        self.item_name = self.items_by_rank[item_rank]() if not name else name
        super().__init__(self.cr, room_name, self.item_name)
        self.treasure = Loot()
        self.treasure.magic_items.append(f"Sentient {self.item_name}")

    def __repr__(self):
        self.output = (
            f"Animated Item: {self.item_name}, CR {cr_str(self.cr_val)}",
            # f"{self.ability_scores}",
            f"Hit Points: {self.total_hp}",
            f"Armor Class: {self.ac}",
            f"Attack Bonus: {self.att_bonus}",
            f"Attack Damage: {self.at_will_dam}",
            f"XP Value: {self.xp}",
            f"Treasure: {self.treasure}",
            ""
        )
        return "\n".join(self.output)


class Villain(Monster):
    special_abilities = TruffleShuffle((
        "Oracle of War. The villain fights with an unnatural defensive advantage.",
        lambda: f"Diabolic Genius. The villain fights with an unnatural "
        f"offensive advantage for the next {dice(2, 2)} turns.",
        "Shadow Step. Once per round as a bonus action, the villain can instantly teleport up to 30 feet.",
        "Blood Pact. The villain will heal to full health.",
        lambda: f"Whirlwind of {damage_types('Physical')}.",
        "Necromancy. The villain will attempt to raise undead champions to fight the party.",
        lambda: f"Breath Weapon. Cone of {damage_types('Arcane')}.",
        "Stone Gaze. The villain will attempt to turn an opponent in to stone.",
        lambda: f"Horrific Intent. This villain will attempt to cause intense fear, preferring to target casters. "
        f"A horrified victim is paralyzed with fear for up to {dice(3, 3)} rounds. "
        f"While horrified, a victim develops a speech impediment and can do little more than wet themselves. "
        f"The paralysis can be negated if an non-fearful ally takes one full turn to snap them out of it. "
        "Even so, the victim will be fearful and attack with disadvantage until the Horrify duration ends.",
        lambda: f"Puppet Master. The villain will attempt to control an opponent. "
        f"The target's body will obey the master in every way for {dice(2, 2)} turns.",
        lambda: f"Raging Madness. The villain grows bigger and more chaotic for {dice(2, 2)} turns. "
        f"While this ability is active the villain gains +5 Attack Roll bonus and -5 AC penalty.",
        lambda: f"Demonic Metamorphosis. The villain transforms into its true form for {dice(2, 2)} turns. "
        "Demonic Metamorphosis is required for this villain to use its special attack. "
        "With this ability active the villain can make 2 special attacks per turn.",
        "Magical Shield. Immune to one damage type of choice for 3 rounds.",
        lambda: f"Mind Games. A random opponent is forced to solve a mental puzzle and is stunned "
        f"for {dice(2, 2)} rounds. Some say this is like playing a game with God, "
        f"but she's playing chess and you're playing checkers. And still, you have no idea why you keep loosing.",
        "Legendary Resistance. If the villain fails a saving throw, they can choose to succeed instead.",
        "Magic Resistance. The villain has advantage on saving throws against spells and magical effects.",
        "Reflective Carapace. Any time the villain is targeted by a magic missile, or any spell that "
        "requires a ranged attack roll, roll a d6. On a roll of 1 to 5, "
        "the villain is unaffected, and the spell is safely deflected away. "
        "On a roll of 6, the villain is still unaffected, but the spell is reflected back at the caster.",
        lambda: f"Natural Damage Resistances: {', '.join(set(damage_types() for _ in range(3)))}."
    ))
    special_weaknesses = TruffleShuffle((
        lambda: f"Susceptible to {damage_types()} damage.",
        "A hidden object in this room holds the villain's soul.",
        lambda: f"The villain is weakened in the presence of a particular magic item. {magic_table_f()}.",
        "The villain is weakened if the entire party laughs at the villain for a full round.",
        "Weapons found in this dungeon deal double damage when used against this villain.",
        "The villain is weakened if its true name is spoken aloud within the villain's range of hearing.",
        "An unknowable prophecy reveals how the villain can be defeated.",
        "An unsolvable riddle reveals how the villain can be defeated.",
        "The villain falls when an ancient enemy forgives them.",
        "The villain loses its special ability if a mystic bargain it struck long ago is satisfied.",
        "The villain is afraid of magic and has disadvantage on saving throws versus spells.",
        lambda: f"Susceptible to {damage_types()} damage.",
    ))

    def __init__(self, cr, room_name="Any", name=None):
        super().__init__(cr, room_name, name)
        self.name = random_monster("villain") if not name else name
        self.dam_type = self.find_dam_type()
        self.treasure = get_loot(self.cr_val, monster_name=self.name, room_name=room_name, is_villain=True)
        self.special = self.special_abilities()
        self.weakness = self.special_weaknesses()
        self.at_will_dam = f"{smart_clamp(self.cr_val // 3, 1, 10)}d8 + {self.prof_bonus} {self.dam_type}"
        self.total_specials = 3
        self.stats = {
            ability: ability_dice(6) for ability in self.abilities
        }
        self.stats[random_value(list(self.stats.keys()))] += dice(2, 4)
        for key, val in self.stats.items():
            self.stats[key] = smart_clamp(self.stats[key], 3, 24)
        self.set_stat_mods()
        self.special_damage = f"{self.damage} {damage_types()}"

    def __repr__(self):
        self.output = (
            f"Villain: {self.name}, CR {cr_str(self.cr_val)}",
            f"Motivation: {self.motive}",
            f"{self.ability_scores}",
            f"Proficiency Bonus: {self.prof_bonus}",
            f"Hit Points: {self.total_hp}",
            f"Armor Class: {self.ac}",
            f"Attack Bonus: {self.att_bonus}",
            f"Number of Attacks: {self.total_specials - 1} basic attacks or 1 special per turn.",
            f"Basic Attack Damage: {self.at_will_dam}",
            f"Special Attack Damage: {self.special_damage}",
            f"Special Ability: {self.special} Usable {self.total_specials} times per day.",
            f"Weakness: {self.weakness}",
            f"Save DC: {self.save_dc}",
            f"XP Value: {self.xp}",
            f"Treasure: {self.treasure}",
            ""
        )
        return '\n'.join(self.output)


class Boss(Villain):
    def __init__(self, cr, room_name="Any", name=None):
        super().__init__(cr, room_name, name=None)
        self.name = random_monster("boss") if not name else name
        self.dam_type = self.find_dam_type()
        while (len(self.treasure) - self.num_potions()) < 3:
            self.treasure += no_coin(get_loot(cr, monster_name=self.name, room_name="Lair", is_villain=True))
        self.total_specials = 5
        if "Legendary" in self.name:
            self.treasure += get_loot(cr, monster_name=self.name, room_name=room_name, is_villain=True)
            second_ability = self.special_abilities()
            while second_ability is self.special:
                second_ability = self.special_abilities()
            self.special += f"\nSpecial Ability: {second_ability}"
        self.at_will_dam = f"{smart_clamp(self.cr_val // 3, 1, 10)}d10 + {self.prof_bonus} {self.dam_type}"
        self.stats = {
            ability: ability_dice(7) for ability in self.abilities
        }
        for _ in range(3):
            self.stats[random_value(list(self.stats.keys()))] += dice(2, 6)
        for key, val in self.stats.items():
            self.stats[key] = smart_clamp(self.stats[key], 3, 30)
        self.set_stat_mods()
        self.special_damage = f"{self.damage} {damage_types()}"

    def num_potions(self):
        return str(self.treasure).count("Potion")

    def __repr__(self):
        self.output = (
            f"Boss: {self.name}, CR {cr_str(self.cr_val)}",
            f"{self.ability_scores}",
            f"Hit Points: {self.total_hp}",
            f"Armor Class: {self.ac}",
            f"Attack Bonus: {self.att_bonus}",
            f"Number of Attacks: {self.total_specials - 1} basic attacks or 1 special per turn.",
            f"Basic Attack: {self.at_will_dam}",
            f"Special Attack: {self.special_damage}",
            f"Special Ability: {self.special}",
            "Legendary Actions. The boss may use one basic attack or move action at the end of every opponent's turn.",
            f"Save DC: {self.save_dc}",
            f"XP Value: {self.xp}",
            f"Treasure: {self.treasure}",
            ""
        )
        return '\n'.join(self.output)


class CampaignBoss(Boss):

    def __init__(self, cr, name=None):
        n = name if name else random_monster("campaign boss")
        super().__init__(cr, name=n)
        while (len(self.treasure) - self.num_potions()) < self.tier:
            self.treasure += no_coin(get_loot(cr, monster_name=self.name, room_name="Lair", is_villain=True))

    def __repr__(self):
        self.output = (
            f"Campaign Boss: {self.name}, CR {cr_str(self.cr_val)}",
            f"{self.ability_scores}",
            f"Hit Points: {self.total_hp}",
            f"Armor Class: {self.ac}",
            f"Attack Bonus: {self.att_bonus}",
            f"Number of Attacks: {self.total_specials} basic attacks or 1 special per turn.",
            f"Basic Attack: {self.at_will_dam}",
            f"Special Attack: {self.special_damage}",
            f"Special Ability: {self.special}",
            "Legendary Actions. The boss may use one basic attack or move action at the end of every opponent's turn.",
            f"Save DC: {self.save_dc}",
            f"XP Value: {self.xp}",
            f"Treasure: {self.treasure}",
            ""
        )
        return '\n'.join(self.output)


if __name__ == "__main__":
    print()
    # print(CampaignBoss(9, name="Black Blade Baron"))
    print(ItemEncounter(10))



from Fortuna import *
from DungeonGenerator.treasure import magic_table_i, magic_table_h


damage_type_cat = {
    "hacking": (
        "Hacking", "Axe", "Broken", "Horned One", "Battleaxe", "Greataxe", "Zweihänder",
        "Greatsword", "Halberd", "Scythe"
    ),
    "slashing": (
        "Slashing", "Blade", "Razor", "Glaive", "Sword", "Scimitar", "Longsword", "Falchion",
        "Gladius", "Saber", "Katana", "Poleaxe", "Whip", "Cleaver", "Bullwhip", "Boline", "Shortsword"
    ),
    "piercing": (
        "Piercing", "Hook", "Spike", "Barb", "Archer", "Ranger", "Horn", "Knife", "Knives", "Pick",
        "Javelin", "Dagger", "Longbow", "Crossbow", "Shortbow", "Spear", "Pilum", "Athame",
        "Sickle", "Rapier", "Lance", "Trident", "Dart", "Pike", "Rapier", "Pitchfork", "Shank", "Stake"
    ),
    "bludgeoning": (
        "Bludgeoning", "Hammer", "Fist", "Defender", "Guardian", "Mace", "Warhammer", "Sling",
        "Morningstar", "Flail", "Scepter", "Greatclub", "Club", "Quarterstaff", "Maul",
        "Shovel", "Tenderizer", "Pipe", "Bookends", "Rolling Pin", "Unarmed", "Cane"
    ),
    "fire": (
        "Fire", "Flame", "Magma", "Heat", "Red", "Ruby", "Burning", "Smoldering",
        "Lava", "Flaming", "Brazier",
    ),
    "frost": ("Frost", "Frozen", "Cold", "Ice", "Snow", "Water", "White", "Diamond", "Yeti"),
    "lightning": ("Lightning", "Electric", "Storm", "Blue", "Sapphire", "Cloud", "Thunder"),
    "acid": ("Acid", "Corrosion", "Bile", "Green", "Emerald"),
    "poison": ("Poison", "Toxic", "Assassin", "Viper"),
    "shadow": ("Shadow", "Darkness", "Negative", "Black", "Onyx", "Dark", "Displacer Beast"),
    "radiant": ("Radiant", "Sun", "Light", "Life", "Gold"),
    "unholy": ("Unholy", "Fallen", "Voodoo", "Devil", "Dark", "Grave", "Pestilence"),
    "necrotic": (
        "Necrotic", "Mummy", "Lich", "Demilich", "Zombie", "Vampire", "Banshee", "Ghost", "Reaper",
        "Corrupted", "Hell", "Necro", "Death", "Infect", "Corrupt", "Nightmare",
    ),
    "nature": ("Nature", "Druid", "Tree", "Rose"),
    "holy light": ("Holy", "Angel", "Celestial", "Heaven"),
    "crushing": ("Crushing", "Behemoth", "Giant", "Bone", "Shield"),
    "Physical": ("Fighter", "Barbarian", "Gladiator", "Rogue", "Pirate", "Beast", "Acrobat", "Valkyrie"),
    "Arcane": (
        "Arcane", "Wizard", "Warlock", "Sorcerer", "Sorceress", "Mage", "Enchantress", "Enchanter",
        "Alchemist", "Shaman", "Witch", "Tome", "Wand", "Rod", "Staff"
    ),
    "Divine": ("Cleric", "Priest", "Priestess", "Paladin", "Zealot", "Cult", "Acolyte", "Divine", "Inquisitor")
}

motivation = TruffleShuffle((
    'Food',
    'Pride',
    'Revenge',
    'Anxiety',
    'Madness',
    'Jealousy',
    'Greed',
    'Survival',
    'Freedom',
    'Hatred and malice',
    'Delusions of grandeur',
    'Trying to take over the world',
    'Self loathing',
    'Trying to conquer the dungeon',
    lambda: f'Looking for a missing {"item: " + magic_table_h() if percent_true(50) else "NPC"}',
    'Trying to slay a rival',
    'Hiding from enemies',
    'Recovering from battle',
    'Avoiding danger',
    'Seeking wealth',
    'Defending its lair',
    'Trying to hibernate',
    'Setting a trap for the party',
    'Trying to remember a magic word',
    'Fear of superficial differences',
    lambda: f'{"Paid" if percent_true(50) else "Coerced"} to assassinate the party',
    'Preparing to ambush the party',
    lambda: f'Trying to protect an {"item: " + magic_table_i() if percent_true(25) else "NPC"}',
    lambda: f'Trying to contact a deceased {"family member" if percent_true(30) else "lover"}',
    "Recently and almost by chance this creature suddenly realized that life is just a game of dice "
    "and decided to go on a murderous rampage starting with you.",
    'Under the effects of a charm spell or love potion.',
    'Still trying to makeup for past choices and come to terms with this whole "be an adult" thing.',
    'Schizophrenia with suicidal tenancies and a twist of too much testosterone, mix well serve with ice.',
    'General paranoia and mild depression, no family, symptoms exhibited during the study include '
    'sporadic outbursts of senseless violence and random acts of brilliant poetry to no one listening. '
    'Patient often refuses medication and has shown no desire to cooperate. Perfect candidate for the program!',
    'Obsessed with puppet shows about cats',
    'Conviction in a false or dieing faith',
    'Brainwashed to believe the media is the enemy and foreigners are evil',
    'Feelings of impedance or inadequacy among peers',
    'Sexual frustration',
    'Rare mating ritual',
    "Trying to impress an unavailable parental figure who can't be bothered to pay attention to anything but "
    "ogre ball and that damn political puppet show.",
))

rand_elemental = TruffleShuffle((
    "Fire",
    "Water",
    "Lightning",
    "Smoke",
    "Mud",
    "Steam",
    "Ice",
    "Magma",
    "Dust",
))

rand_color = FlexCat({
    "chromatic": (
        "Red",
        "Green",
        "Blue",
        "Black",
        "White",
        "Shadow",
    ),
    "crystal": (
        "Ruby",
        "Emerald",
        "Sapphire",
        "Quartz",
        "Diamond",
        "Onyx",
    ),
    "metal": (
        "Brass",
        "Copper",
        "Bronze",
        "Silver",
        "Gold",
        "Platinum",
    )
}, key_bias="front_gauss", val_bias="front_linear")

rand_knight = TruffleShuffle((
    "Flaming Tongue",
    "Frozen Waste",
    "Drunken Monkey",
    "Hidden Rose",
    "Thunderbolts",
    "Shadows",
    "Broken Lance",
    "Grim Reaper",
    "Three-eyed Raven",
))

rand_adventurer = TruffleShuffle((
    lambda: f"Fighter",
    lambda: f"Rogue",
    lambda: f"Beast Master",
    lambda: f"Archer",
    lambda: f"Cleric",
    lambda: f"Wizard",
    lambda: f"Warlock" if percent_true(50) else f"Witch",
    lambda: f"Sorceress" if percent_true(50) else f"Sorcerer",
    lambda: f"Barbarian" if percent_true(50) else f"Valkyrie",
    lambda: f"Dark Paladin",
    lambda: f"Demonic Summoner",
    lambda: f"Enchantress" if percent_true(50) else f"Enchanter",
    lambda: f"Voodoo Witch Doctor",
    lambda: f"Wild Mage",
    lambda: f"Knight of the {rand_knight()}",
    lambda: f"Defender",
    lambda: f"Acolyte",
    lambda: f"Priestess" if percent_true(50) else f"Priest",
    lambda: f"Zealot",
    lambda: f"Cultist",
    lambda: f"Gladiator",
    lambda: f"Pirate",
    lambda: f"Ninja",
    lambda: f"Alchemist",
    lambda: f"Artificer",
    lambda: f"Shaman",
    lambda: f"Blood Witch",
    lambda: f"Necromancer",
    lambda: f"Arcane Trickster",
    lambda: f"Acrobat",
    lambda: f"Assassin",
    lambda: f"Inquisitor",
))


rand_giant = TruffleShuffle((
    'Hill',
    'Stone',
    'Cloud',
    'Frost',
    'Fire',
    'Storm',
    'Two-headed',
))

rand_race = QuantumMonty((
    "Human",
    "Dwarf",
    "Halfling",
    "Gnome",
    "Elf",
    "Half-elf",
    "Drow",
    "Tiefling",
    "Dragonborn",
    "Half-orc",
    "Goblin",
    "Orc",
    "Half-giant",
    "Half-dragon",
)).front_linear


rand_devil = TruffleShuffle((
    "Flames",
    "Ice",
    "Barbs",
    "Bones",
    "Chains",
    "Blades",
    "Hooks",
    "the Horned One",
))

rand_lycanthrope = QuantumMonty((
    "Werewolf",
    "Wererat",
    "Werecat",
    "Wereboar",
    "Werebear",
    "Werecheetah",
    "Werepanther",
    "Weretiger",
    "Werelion",
)).front_linear

hydra_heads = QuantumMonty((
    "Three",
    "Four",
    "five",
    "six",
    "seven",
    "eight",
    "nine",
))

random_monster = FlexCat({
    'minion': (
        lambda: f"Novice {rand_adventurer()}",
        lambda: f"{rand_elemental()} Mephit",
        lambda: f"Zombie {rand_race()}",
        lambda: f"Troll",
        lambda: f"Ogre",
        lambda: f"Jackalope",
        lambda: f"Imp",
        lambda: f"Skeleton {rand_adventurer()}",
        lambda: f"Quasit",
        lambda: f"Gremlin",
        lambda: f"Kobold",
        lambda: f"Gnoll",
        lambda: f"Gelatinous Cube",
        lambda: f"Drider",
        lambda: f"Tasmanian Devil",
        lambda: f"Crawling Claw",
        lambda: f"Vampire Spawn, {rand_race()}",
        lambda: f"Gargoyle",
        lambda: f"Hell Hound",
        lambda: f"Gelatinous Spores",
        lambda: f"Sasquatch",
        lambda: f"Yeti",
        lambda: f"{rand_color()} Dragon Hatchling",
    ),
    'monster': (
        lambda: f"Goblin {rand_adventurer()}",
        lambda: f"{rand_elemental()} Elemental",
        lambda: f"{rand_giant()} Giant",
        lambda: f"Chimera",
        lambda: f"Demon Cat",
        lambda: f"Basilisk",
        lambda: f"Lycanthrope: {rand_lycanthrope()}",
        lambda: f"Hobgoblin",
        lambda: f"Cyclops",
        lambda: f"Umber Hulk",
        lambda: f"Bugbear",
        lambda: f"Owlbear",
        lambda: f"Winged Kobold",
        lambda: f"Cave Troll",
        lambda: f"Ghost: {rand_race()}",
        lambda: f"Displacer Beast",
        lambda: f"Gelatinous Sphere",
        lambda: f"Golem",
        lambda: f"Shield Guardian",
        lambda: f"Wyvern",
        lambda: f"Wraith",
        lambda: f"Intellect Devourer",
        lambda: f"Ghoul",
        lambda: f"Shadow",
        lambda: f"Rust Monster",
        lambda: f"Vampire, {rand_race()}",
        lambda: f"Worg",
        lambda: f"Hook Horror",
        lambda: f"Mimic",
        lambda: f"Flameskull",
        lambda: f"Xorn",
        lambda: f"Bafu",
        lambda: f"Chupacabra",
        lambda: f"Skeletal Monstrosity",
        lambda: f"Young {rand_color()} Dragon",
        lambda: f"Nightmare",
        lambda: f"Bile Demon",
    ),
    'villain': (
        lambda: f"{rand_race()} {rand_adventurer()}",
        lambda: f"{rand_elemental()} Demon",
        lambda: f"Incubus" if percent_true(30) else f"Succubus",
        lambda: f"Doppelganger: {rand_race()} {rand_adventurer()}",
        lambda: f"Wailing Banshee",
        lambda: f"{'Goblin' if percent_true(50) else 'Gnomish'} War Tank",
        lambda: f"Devil of {rand_devil()}",
        lambda: f"Legendary Lycanthrope: {rand_lycanthrope()} {rand_adventurer()}",
        lambda: f"Efreeti",
        lambda: f"Mummy",
        lambda: f"Elder Vampire, {rand_race()}",
        lambda: f"Death Tyrant",
        lambda: f"Demilich",
        lambda: f"Mind Flayer",
        lambda: f"Adult {rand_color()} Dragon",
        lambda: f"Legendary {rand_race()} {rand_adventurer()}",
    ),
    'boss': (
        lambda: f"Hydra with {hydra_heads()} heads",
        lambda: f"The Rat King",
        lambda: f"Behemoth",
        lambda: f"Cerberus",
        lambda: f"Arch Devil of {rand_devil()}",
        lambda: f"Poltergeist",
        lambda: f"Death Knight",
        lambda: f"Pit Lord",
        lambda: f"Vampire Lord, {rand_race()}",
        lambda: f"Mummy Lord",
        lambda: f"Beholder",
        lambda: f"Ancient {rand_color()} Dragon",
        lambda: f"Ancient Dracolich",
    ),
    'campaign boss': (
        lambda: f"Elemental Lord of {rand_elemental()}",
        lambda: f"Goblin King",
        lambda: f"Balor",
        lambda: f"Legendary Devil of {rand_devil()}",
        lambda: f"Mummy Pharaoh",
        lambda: f"Lich King",
        lambda: f"Beholder Overseer",
        lambda: f"Legendary Vampire: {rand_race()} {rand_adventurer()}",
        lambda: f"Lord of the Pit",
        lambda: f"Grim Reaper",
        lambda: f"Flying Spaghetti Monster",
        lambda: f"Legendary {rand_color()} Dragon",
        lambda: f"The Nameless One",
    ),
})


random_undead = FlexCat({
    'minion': (
        lambda: f"Grave Robber",
        lambda: f"Zombie {rand_race()}",
        lambda: f"Crawling Claw",
        lambda: f"Vampire Spawn {rand_race()}",
        lambda: f"Gargoyle",
        lambda: f"Hell Hound",
        lambda: f"Mimic",
        lambda: f"Shadow",
        lambda: f"Pestilence Rat",
        lambda: f"Skeleton {rand_adventurer()}",
    ),
    'monster': (
        lambda: f"Ghost: {rand_race()} {rand_adventurer()}",
        lambda: f"Displacer Beast",
        lambda: f"Wraith",
        lambda: f"Ghoul",
        lambda: f"Shadow",
        lambda: f"Worg",
        lambda: f"Demon Cat",
        lambda: f"Skeletal Monstrosity",
        lambda: f"Vampire: {rand_race()}",
    ),
    'villain': (
        lambda: f"Doppelganger: {rand_race()} {rand_adventurer()}",
        lambda: f"Legendary Ghost: {rand_race()} {rand_adventurer()}",
        lambda: f"Corrupted Nun",
        lambda: f"Dark Summoner",
        lambda: f"Banshee",
        lambda: f"Nightmare",
        lambda: f"Mummy",
        lambda: f"Hook Horror",
        lambda: f"Flameskull",
        lambda: f"Grave Dust Elemental",
        lambda: f"Shadow Elemental",
        lambda: f"Lich",
        lambda: f"Elder Vampire, {rand_race()}",
    ),
    'boss': (
        lambda: f"Death Tyrant",
        lambda: f"Demilich",
        lambda: f"Poltergeist",
        lambda: f"Death Knight",
        lambda: f"Shadow Fiend",
        lambda: f"Dracolich",
        lambda: f"Mummy Lord",
        lambda: f"Vampire Lord {rand_race()}",
    ),
    'campaign boss': (
        lambda: f"Legendary Devil of {rand_devil()}",
        lambda: f"Mummy Pharaoh",
        lambda: f"Lich King",
        lambda: f"Lord of the Pit",
        lambda: f"Grim Reaper",
        lambda: f"Shadow Demon",
        lambda: f"Shadow Dragon",
        lambda: f"Legendary Vampire: {rand_race()} {rand_adventurer()}",
    ),
})


corporeal_undead = FlexCat({
    'minion': (
        "Zombie",
        "Crawling Claw",
        "Vampire Spawn",
        "Pestilence Rat",
        "Skeleton",
    ),
    'monster': (
        "Flameskull",
        "Ghoul",
        "Frozen Walker",
        "Skeletal Monstrosity",
        "Fledgling Vampire",
    ),
    'villain': (
        "Mummy",
        "Lich",
        "Vampire",
        "Mummy Lord",
        "Death Knight",
    ),
    'boss': (
        "Death Tyrant",
        "Ancient Vampire",
        "Demilich",
    ),
    'campaign boss': (
        "Mummy Pharaoh",
        "Lich King",
        "Dracolich",
    )
})


incorporeal_undead = FlexCat({
    'minion': (
        "Ghostly Villager",
        "Crawling Mist",
        "Willow-the-wisp",
        "Specter",
    ),
    'monster': (
        "Spectral Beast",
        "Lost Soul",
        "Weeping Apparition",
        "Ghost",
    ),
    'villain': (
        "Tortured Spirit",
        "Shade",
        "Wraith",
        "Banshee",
        "Spectral Giant",
    ),
    'boss': (
        "Wight",
        "Nightmare",
        "Haunted Horror",
    ),
    'campaign boss': (
        "Soul Reaper",
        "Poltergeist",
        "Spectral Dragon",
    ),
}, val_bias="front_linear")



from Fortuna import *


def cr_adapter(num_players, average_level, difficulty=0):
    average_level = smart_clamp(average_level, 1, 20)
    num_players = smart_clamp(num_players, 1, 9) - 5
    magic_bonus = (0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4)[average_level]
    diff_modifier = smart_clamp(difficulty, -5, 5)
    return smart_clamp(num_players + average_level + magic_bonus + diff_modifier, -3, 30)


def cr_str(cr):
    if type(cr) == int:
        return ("1/16", "1/8", "1/4", "1/2")[cr + 3] if cr < 1 else str(cr)
    else:
        return cr.string


damage_types = FlexCat({
    "Physical": ('slashing', 'bludgeoning', 'piercing', 'hacking', 'crushing'),
    "Arcane": ('fire', 'lightning', 'frost', 'acid', 'shadow', 'radiant'),
    "Unholy": ('necrotic', 'hell fire', 'corruption', 'void', 'negative energy'),
    "Holy": ('holy fire', 'thunderbolt', 'holy light', 'divine force', 'positive energy'),
    "Nature": (lambda: damage_types("Arcane"), lambda: damage_types("Physical")),
    "Divine": (lambda: damage_types("Holy"), lambda: damage_types("Nature"), lambda: damage_types("Unholy")),
}, key_bias="front_linear", val_bias="truffle_shuffle")


if __name__ == "__main__":
    print()
    for _ in range(20):
        print(damage_types())



from DungeonGenerator.npc_lib import npc_dict


things_dict = {
    'actions': npc_dict["quests"],
    'trinkets': [f"Find {itm}" for itm in npc_dict["trinkets"]],
    "skills": (
        "Acrobatics, DEX",
        "Animal handling, WIS",
        "Arcana, INT",
        "Athletics, STR",
        "Deception, CHA",
        "History, INT",
        "Insight, WIS",
        "Intimidation, CHA",
        "Investigation, INT",
        "Medicine, WIS",
        "Nature, INT",
        "Perception, WIS",
        "Performance, CHA",
        "Persuasion, CHA",
        "Religion, INT",
        "Sleight of hand, DEX",
        "Stealth, DEX",
        "Survival, WIS"
    )
}



from Fortuna import *
from DungeonGenerator.cr import CR
from DungeonGenerator.gear_lib import basic_gear


class BasicGear:
    gear_types = basic_gear.random_cat

    def __init__(self, cr):
        self.cr = smart_clamp(cr, -3, 30)
        self.id, self.item_type, self.price, self.cursed, self.magic = self.make_item(self.cr)
        self.name = f"{self.item_type}, Price: {self.price}"

    def __repr__(self):
        return self.name

    def identify(self):
        self.name = self.id

    def make_item(self, cr):
        item_type = self.gear_types()
        is_cursed = percent_true(cr / 3)
        if cr < 1:
            price = f"{50 + (cr * 10) + d(10)} Copper"
            magic_value = 0
        elif cr <= 4:
            price = f"{cr * 50 + d(100)} Copper"
            magic_value = 0
        elif cr <= 8:
            price = f"{cr * 50 + d(100)} Silver"
            if is_cursed:
                magic_value = -1
            else:
                magic_value = d(2) - 1
        elif cr <= 12:
            price = f"{cr * 50 + d(100)} Electrum"
            if is_cursed:
                magic_value = -d(2)
            else:
                magic_value = d(2)
        elif cr <= 16:
            price = f"{cr * 50 + d(100)} Gold"
            if is_cursed:
                magic_value = -d(2)
            else:
                magic_value = d(2)
        else:
            price = f"{cr * 50 + d(100)} Platinum"
            if is_cursed:
                magic_value = -d(3)
            else:
                magic_value = d(3)
        if magic_value == 0 or "Clothing" in item_type:
            item_good = percent_true(50)
            if item_good:
                item_quality = "exceptional"
            else:
                item_quality = "poor"
            item = f"{basic_gear(item_type)}, {item_quality} quality"
        elif magic_value > 0:
            item = f"{basic_gear(item_type)}, +{magic_value}"
        else:
            item = f"{basic_gear(item_type)}, {magic_value} Cursed"
        return item, item_type, price, is_cursed, magic_value


def random_gear(cr, identify=False):
    item = BasicGear(cr)
    if identify:
        item.identify()
    return item


class Potion:
    random_potion = FlexCat({
        1: (
            "Potion case: 10 Potions of healing",
            "Potion of animal friendship",
            "Potion of water breathing",
            "Potion of hill giant strength",
            "Potion of cloud giant strength",
            "Potion of climbing",
            "Potion of poison",
            "Potion of speed",
            "Potion of growth",
            "Potion of fire breath",
            "Potion of diminution",
            "Potion of gaseous form",
        ),
        2: (
            "Potion case: 10 Potions of greater healing",
            "Potion of frost giant strength",
            "Potion of fire giant strength",
            "Potion of clairvoyance",
            "Potion of vitality",
            "Potion of flying",
            "Potion of invisibility",
            "Potion of resistance",
        ),
        3: (
            "Potion case: 10 Potions of superior healing",
            "Potion of stone giant strength",
            "Potion of mind reading",
            "Potion of heroism",
        ),
        4: (
            "Potion case: 10 Potions of supreme healing",
            "Potion of storm giant strength",
            "Potion of longevity",
            "Potion of invulnerability",
        ),
    }, key_bias="front_linear", val_bias="truffle_shuffle")

    def __init__(self, cr):
        tier = d(CR(cr).tier)
        self.name = self.random_potion(tier)
        self.price = f"{dice(5, 100)} {('Silver', 'Electrum', 'Gold', 'Platinum')[tier - 1]}"

    def __repr__(self):
        return f"{self.name}, Price: {self.price}"


if __name__ == '__main__':
    print()
    for _ in range(10):
        p = BasicGear(5)
        print(f"{p}, {p.id}")



from Fortuna import smart_clamp


class CR:
    """ The CR class is a numeric system representing the relative power of a monster in D&D 5e.
    This system is a bit funky with values below 1, be careful... here be dragons!
    CR less than 1 are printed as fractions but valued mathematically as integers [-3, 0].
    CR(-3) is CR 1/16 - same as CR 0 in the books.
    CR(-2) is CR 1/8
    CR(-1) is CR 1/4
    CR(0) is CR 1/2
    CR(1)..CR(30) is CR 1 to CR 30
    Why is this system so crazy? Because wizards are bad at math!
    And I wanted to ignore the fractions...
    So that subtracting 1 always takes you to the next CR down.
    Currently: CR(1) + CR(1) is CR(2)
    what is CR 1/2 + CR 1/4 ???
    This is an open question! It may cause me to switch to floats internally, eww.
    """

    def __init__(self, val):
        self.val = smart_clamp(val, -3, 30)

    @property
    def value(self):
        return self.val

    @property
    def key(self):
        return self.val + 3

    @property
    def string(self):
        if self.val > 0:
            str_cr = f"{self.val}"
        else:
            str_cr = ("1/16", "1/8", "1/4", "1/2")[self.key]
        return f"CR {str_cr}"

    @property
    def tier(self):
        return (
            1, 1, 1, 1,
            1, 1, 1, 1, 1,
            2, 2, 2, 2, 2,
            3, 3, 3, 3, 3,
            4, 4, 4, 4, 4,
            4, 4, 4, 4, 4,
            4, 4, 4, 4, 4,
        )[self.key]

    @classmethod
    def party_adapter(cls, average_level, num_players=5, difficulty=0):
        average_level = smart_clamp(average_level, 1, 20)
        num_players = smart_clamp(num_players, 1, 9) - 5
        magic_bonus = (0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4)[average_level]
        diff_modifier = smart_clamp(difficulty, -5, 5)
        return CR(num_players + average_level + magic_bonus + diff_modifier)

    def __repr__(self):
        return self.string

    def __add__(self, other):
        self.val = smart_clamp(self.val + other, -3, 30)
        return self

    def __sub__(self, other):
        self.val = smart_clamp(self.val - other, -3, 30)
        return self


if __name__ == "__main__":
    for i in range(-3, 31):
        cr = CR(i)
        print(f"{cr}, Tier: {cr.tier}")



from setuptools import setup


with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="DungeonGenerator",
    author="Robert Sharp",
    author_email="webmaster@sharpdesigndigital.com",
    install_requires=["Fortuna"],
    packages=["DungeonGenerator"],
    version="0.1.9",
    description="D&D Dungeon Generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Free for non-commercial use",
    platforms=["Darwin", "Linux"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=["D&D", "Dungeon", "Monster", "Treasure", "Traps"],
    python_requires='>=3.6',
)
